/*   limits.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "externsd.h"
#include "log.h"
#include "serverd.h"
#include "user.h"
#include "utils.h"

/* Syntax:
   <entry>     := <days> <time> <users> <bandwidth>
   <days>      := <weekday>-<weekday> |
                  * | ~
   <time>      := <hour>:<minute>-<hour>:<minute> |
                  * | ~
   <users>     := [u=<n>][,a=<n>][,i=<n>,d=<n>] |
                  ua=<n>[,i=<n>,d=<n>] |
                  * | ~
   <bandwidth> := [u=<n>[/<r>]][,d=<n>[/<r>]]  |
                  ud=<n>[/<r>] |
                  * | ~
*/

int             bw_dl       = 0;      /* handed out dl bandwidth */
int             bw_ul       = 0;      /* handed out ul bandwidth */
struct limit_s* limit_array = NULL;   /* limits                  */
struct limit_s* limit       = NULL;   /* current limit           */
int             limit_cnt   = 0;      /* number of limits        */
int             limit_max   = 0;      /* size of limit_array     */

int lim_in_timerange(struct limit_s* l, struct tm* t)
{ int intime;

  if (l->sday <= l->eday)
  { if ((t->tm_wday < l->sday) || (t->tm_wday > l->eday)) return 0;
  } else
  { if ((t->tm_wday < l->sday) && (t->tm_wday > l->eday)) return 0;
  }
  if (l->shour < l->ehour) intime = 1; else
    if (l->shour > l->ehour) intime = 0; else
    intime = (l->smin <= l->emin);
  if (intime)
  { if ((t->tm_hour < l->shour) || (t->tm_hour > l->ehour)) return 0;
  } else
  { if ((t->tm_hour < l->shour) && (t->tm_hour > l->ehour)) return 0;
  }
  if (t->tm_hour == l->shour) if (t->tm_min < l->smin) return 0;
  if (t->tm_hour == l->ehour) if (t->tm_min > l->emin) return 0;
  return 1;
}

void lim_update(time_t tim)
{ int        i;
  struct tm* t;

  if (!limit_cnt) limit = NULL; else
  { t = localtime(&tim);
    for (i=0, limit=limit_array; i<limit_cnt; i++, limit++)
    { if (lim_in_timerange(limit,t)) break;
    }
    if (i >= limit_cnt) limit = NULL;
  }
}

int lim_alloc(int n)
{ int   new_max;
  void* ptr;

  new_max = limit_max + n;
  ptr = alloc(limit_array,new_max*sizeof(struct limit_s));
  if (!ptr) return limit_max;
  limit_array = ptr;
  limit_max = new_max;
  return limit_max;
}

void lim_default()
{ limit = NULL;
  limit_cnt = 0;
}

char* weekdays[] = { "su", "mo", "tu", "we", "th", "fr", "sa", NULL };

int lim_parse(char* str)
{ char            buf[8];
  int             i;
  struct limit_s* l;
  char*           p;
  int             what;

  if (limit_cnt >= limit_max)
  { lim_alloc(8);
    if (limit_cnt >= limit_max)
    { errlog("out of memory (limits)");
      return 0;
    }
  }
  l = limit_array + limit_cnt;
  p = str;
  
  /* scan days */
  while (isspace(*p)) p++;
  if ((*p == '#') || (*p == ';')) return 0;
  if (*p == '*')
  { p++;
    l->sday = 0;
    l->eday = 6;
  } else if (*p == '~')
  { p++;
    if (l == limit) return -10;
    l->sday = (l-1)->sday;
    l->eday = (l-1)->eday;
  } else
  { buf[2] = '\0';
    strncpy(buf,p,2);
    for (i=0; weekdays[i]; i++) if (!strcasecmp(buf,weekdays[i]))
    { l->sday = i;
      break;
    }
    if (!weekdays[i]) return -11;
    p += 2;
    if (*p == '-')
    { strncpy(buf,++p,2);
      for (i=0; weekdays[i]; i++) if (!strcasecmp(buf,weekdays[i]))
      { l->eday = i;
        break;
      }
      if (!weekdays[i]) return -12;
      p += 2;
    } else l->eday = l->sday;
  }

  /* scan time */
  while (isspace(*p)) p++;
  if (*p == '*')
  { p++;
    l->shour = l->smin = 0;
    l->ehour = 23;
    l->emin = 59;
  } else if (*p == '~')
  { p++;
    if (l == limit) return -20;
    l->shour = (l-1)->shour;
    l->smin = (l-1)->smin;
    l->ehour = (l-1)->ehour;
    l->emin = (l-1)->emin;
  } else
  { if (sscanf(p,"%u:%u",&l->shour,&l->smin) != 2) return -21;
    while (isdigit(*p) || (*p == ':')) p++;
    if (*p == '-')
    { if (sscanf(++p,"%u:%u",&l->ehour,&l->emin) != 2) return -22;
      while (isdigit(*p) || (*p == ':')) p++;
    } else
    { l->ehour = l->shour;
      l->emin = l->smin;
    }
    if ((l->shour > 23) || (l->smin > 59) || (l->ehour > 23) || (l->emin > 59))
      return -23;
  }

  /* scan max users */
  while (isspace(*p)) p++;
  if (*p == '*')
  { p++;
    l->max_anon = l->max_anon_dom = l->max_anon_ip = l->max_user = -1;
    l->max_sum = 0;
  } else if (*p == '~')
  { p++;
    if (l == limit) return -30;
    l->max_anon = (l-1)->max_anon;
    l->max_anon_dom = (l-1)->max_anon_dom;
    l->max_anon_ip = (l-1)->max_anon_ip;
    l->max_user = (l-1)->max_user;
    l->max_sum = (l-1)->max_sum;
  } else
  { l->max_anon = l->max_anon_dom = l->max_anon_ip = l->max_user = -1;
    l->max_sum = 0;
    while (*p && !isspace(*p))
    { if (*p == ',') p++;
      switch (*(p++))
      { case 'a' :
          if (*p == 'u')
          { p++;
            l->max_sum = 1;
            what = 4;
          } else what = 1;
          break;
        case 'd' :
          what = 2;
          break;
        case 'i' :
          what = 3;
          break;
        case 'u' :
          what = 4;
          if (*p == 'a')
          { p++;
            l->max_sum = 1;
          }
          break;
        default  :
          return -31;
      }
      if (*(p++) != '=') return -32;
      i = 0;
      if (*p == '-') buf[i++] = *(p++);
      while (isdigit(*p))
      { buf[i++] = *(p++);
        if (i == sizeof(buf)) return -33;
      }
      buf[i] = '\0';
      if (!sscanf(buf,"%i",&i)) return -34;
      switch (what)
      { case 1 : l->max_anon = i; break;
        case 2 : l->max_anon_dom = i; break;
        case 3 : l->max_anon_ip = i; break;
        case 4 : l->max_user = i; break;
      }
    }
  }

  /* scan bandwidth */
  while (isspace(*p)) p++;
  if (*p == '*')
  { p++;
    l->sp_dl = l->sp_ul = -1;
    l->sp_sum = l->sp_dl_fact = l->sp_ul_fact = 0;
  } else if (*p == '~')
  { p++;
    if (l == limit) return -40;
    l->sp_dl = (l-1)->sp_dl;
    l->sp_ul = (l-1)->sp_ul;
    l->sp_sum = (l-1)->sp_sum;
    l->sp_dl_fact = (l-1)->sp_dl_fact;
    l->sp_ul_fact = (l-1)->sp_ul_fact;
  } else
  { l->sp_dl = l->sp_ul = -1;
    l->sp_sum = l->sp_dl_fact = l->sp_ul_fact = 0;
    while (*p && !isspace(*p))
    { if (*p == ',') p++;
      switch (*(p++))
      { case 'd' :
          if (*p == 'u')
          { p++;
            l->sp_sum = 1;
            what = 0x03;
          } else what = 0x01;
          break;
        case 'u' :
          if (*p == 'd')
          { p++;
            l->sp_sum = 1;
            what = 0x03;
          } else what = 0x02;
          break;
        default  :
          return -41;
      }
      if (*(p++) != '=') return -42;
      i = 0;
      if (*p == '-') buf[i++] = *(p++);
      while (isdigit(*p))
      { buf[i++] = *(p++);
        if (i == sizeof(buf)) return -43;
      }
      buf[i] = '\0';
      if (!sscanf(buf,"%i",&i)) return -44;
      if (what & 0x01) l->sp_dl = i;
      if (what & 0x02) l->sp_ul = i;
      if (*p == '/')
      { p++;
        i = 0;
        while (isdigit(*p))
        { buf[i++] = *(p++);
          if (i == sizeof(buf)) return -45;
        }
        buf[i] = '\0';
        if (!sscanf(buf,"%i",&i)) return -46;
        if (what & 0x01) l->sp_dl_fact = i;
        if (what & 0x02) l->sp_ul_fact = i;
      }
    }
  }

  /* get max number of users */
  i = l->max_sum ? l->max_user : ((l->max_user < 0) || (l->max_anon < 0)) ?
    -1 : l->max_user + l->max_anon;
  /* calculate download speeds */
  if ((l->sp_dl < 0) || !i)
  { l->sp_dl_avg = l->sp_dl_dbl = l->sp_dl_half = -1;
  } else if (l->sp_dl_fact)
  { l->sp_dl_avg = l->sp_dl_dbl = l->sp_dl_half = 0;
  } else if (i < 0)
  { l->sp_dl_avg = 2;
    l->sp_dl_dbl = 4;
    l->sp_dl_half = 1;
  } else
  { l->sp_dl_avg = l->sp_dl / i;
    l->sp_dl_dbl = (l->sp_dl << 1) / i;
    l->sp_dl_half = l->sp_dl / (i << 1);
    if (l->sp_dl_avg <= 0) l->sp_dl_avg = 1;
    if (l->sp_dl_half <= 0) l->sp_dl_half = 1;
    if (l->sp_dl_dbl <= 0) l->sp_dl_dbl = 1;
  }
  /* calculate upload speeds */
  if ((l->sp_ul < 0) || !i)
  { l->sp_ul_avg = l->sp_ul_dbl = l->sp_ul_half = -1;
  } else if (l->sp_ul_fact)
  { l->sp_ul_avg = l->sp_ul_dbl = l->sp_ul_half = 0;
  } else if (i < 0)
  { l->sp_ul_avg = 2;
    l->sp_ul_dbl = 4;
    l->sp_ul_half = 1;
  } else
  { l->sp_ul_avg = l->sp_ul / i;
    l->sp_ul_dbl = (l->sp_ul << 1) / i;
    l->sp_ul_half = l->sp_ul / (i << 1);
    if (l->sp_ul_avg <= 0) l->sp_ul_avg = 1;
    if (l->sp_ul_half <= 0) l->sp_ul_half = 1;
    if (l->sp_ul_dbl <= 0) l->sp_ul_dbl = 1;
  }

  /* all ok */
  limit_cnt++;
  return 1;
}

int lim_read_config(void)
{ char  buf[256];
  FILE* file;
  int   i;

  mk_filename(buf,DFL_LIMITFILE,DFL_LIMITFILE);
  file = fopen(buf,"rt");
  if (!file)
  { errlog("fopen(%s): %s (%i)",buf,strerror(errno),errno);
    lim_default();
    return -1;
  }
  limit_cnt = 0;
  while (readline(buf,sizeof(buf),file))
  { i = lim_parse(buf);
    if (i < 0) errlog("lim_read_config(): syntax error (%i): \"%s\"",i,buf);
  }
  fclose(file);
  lim_update(time(NULL));
  return 0;
}

void lim_return_session(struct child_s* c)
{ switch (c->lim_session)
  { case 1 : num_users--; break;
    case 2 : num_anon--; break;
  }
  c->lim_session = 0;
}

int anon_ip_ok(int ip)
{ struct child_s* ch;
  int             i;
  int             ndom;
  int             nip;

  if (!limit->max_anon_ip || !limit->max_anon_dom) return 0; else
  if ((limit->max_anon_ip > 0) || (limit->max_anon_dom > 0))
  { for (i=0, ndom=0, nip=0, ch=child; i<children; i++, ch++)
    if (anonuser(ch->user))
    { if (ch->ip == ip) nip++;
      if ((ip & 0xFFFFFF00) == (ch->ip & 0xFFFFFF00)) ndom++;
    }
    if (limit->max_anon_ip > 0) if (nip >= limit->max_anon_ip) return 0;
    if (limit->max_anon_dom > 0) if (ndom >= limit->max_anon_dom) return 0;
  }
  return 1;
}

int in_iprange(unsigned long ip, unsigned char* iprange)
{ int n;

  n = (ip >> 24) & 0xFF;
  if ((n < *(iprange)) || (n > *(iprange+1))) return 0;
  n = (ip >> 16) & 0xFF;
  if ((n < *(iprange+2)) || (n > *(iprange+3))) return 0;
  n = (ip >> 8) & 0xFF;
  if ((n < *(iprange+4)) || (n > *(iprange+5))) return 0;
  n = ip & 0xFF;
  if ((n < *(iprange+6)) || (n > *(iprange+7))) return 0;
  return 1;
}

int in_ident(int ip, char* ident, char* iprange)
{ if (in_iprange(ip,(unsigned char*)iprange+12))
  { if (*iprange=='\0' || *iprange=='*' || !flag_get(flags,FLAG_IDENTMODE)) return 1; else
    { if (!strcmp(ident,iprange)) return 1; else return 0;
    }
  }
  return 0;
}

/* returns:
    1 = session ok
    0 = too many users
   -1 = user:logged in max times, anon:ip/domain limit reached
   -2 = bad ident / ip range
*/
int lim_request_session(struct child_s* c, struct user_s* u, int doit)
{ struct child_s* ch;
  int             i;
  int             j;

  if (doit) lim_return_session(c);
  if (anonuser(u))
  { /* ANONYMOUS */
    if (!doit) return 1;
    if (limit)
    { if (limit->max_sum)
      { if ((limit->max_user >= 0) && (num_anon + num_users >= limit->max_user))
          return 0;
      } else
      { if ((limit->max_anon >= 0) && (num_anon >= limit->max_anon))
          return 0;
      }
      if (!anon_ip_ok(c->ip)) return -1;
    }
    if (!c->admin)
    { num_anon++;
      c->lim_session = 2;
    }
  } else
  { /* NORMAL OR SUPERUSER */
    for (i=0; i<10; i++) if (in_ident(c->ip,c->ident,u->idents[i])) break;
    if (i >= 10) return -2;
    if (!doit) return 1;
    if (!superuser(u))
    { if (u->class & UL_MASK)
      { j = (u->class & UL_MASK) >> 8;
        for (i=0, ch=child; i<children; i++, ch++) if (ch->user == u)
          if (!(--j)) return -1;
      }
      if (limit)
      { if (limit->max_sum)
        { if ((limit->max_user >= 0) && (num_anon + num_users >= limit->max_user))
            return 0;
        } else
        { if ((limit->max_user >= 0) && (num_users >= limit->max_user))
            return 0;
        }
      }
    }
    if (!c->admin)
    { num_users++;
      c->lim_session = 1;
    }
  }
  return 1;
}

void lim_return_bandwidth(struct child_s* c)
{ if (c->lim_bw > 0)
  { switch (c->lim_direction)
    { case 0 : bw_dl -= c->lim_bw; break;   /* download */
      case 1 : bw_ul -= c->lim_bw; break;   /* upload   */
    }
    c->lim_bw = 0;
  }
}

int lim_request_bandwidth(struct child_s* c, int direction)
{ int avl;
  int bw;

  lim_return_bandwidth(c);
  switch (direction)
  { case 0  : /* download */
      if (limit && limit->sp_dl >= 0)
      { if (limit->sp_dl)
        { avl = limit->sp_dl - bw_dl;
          if (limit->sp_sum) avl -= bw_ul;
          if (avl > 1)
          { if (limit->sp_dl_fact)
            { bw = avl / limit->sp_dl_fact;
              if (bw < 1) bw = 1;
            } else
            { bw = (avl * 100) / limit->sp_dl;
              if (bw > 50) bw = limit->sp_dl_dbl; else
              if (bw > 30) bw = limit->sp_dl_avg; else
                bw = limit->sp_dl_half;
            }
          } else bw = 1; /* avaliable    <= 1  */
        } else bw = 0;   /* limit->sp_dl == 0  */
      } else bw = -1;    /* limit->sp_dl == -1 */
      avl = c->user->dl_speed >> 16;
      if (avl) if ((bw < 0) || (bw > avl)) bw = avl;
      if (bw > 0) bw_dl += bw;
      break;
    case 1  : /* upload */
      if (limit && limit->sp_ul >= 0)
      { if (limit->sp_ul)
        { avl = limit->sp_ul - bw_ul;
          if (limit->sp_sum) avl -= bw_dl;
          if (avl > 1)
          { if (limit->sp_ul_fact)
            { bw = avl / limit->sp_ul_fact;
              if (bw < 1) bw = 1;
            } else
            { bw = (avl * 100) / limit->sp_ul;
              if (bw > 50) bw = limit->sp_ul_dbl; else
              if (bw > 30) bw = limit->sp_ul_avg; else
                bw = limit->sp_ul_half;
            }
          } else bw = 1; /* avaliable    <= 1  */
        } else bw = 0;   /* limit->sp_ul == 0  */
      } else bw = -1;    /* limit->sp_ul == -1 */
      avl = c->user->ul_speed >> 16;
      if (avl) if ((bw < 0) || (bw > avl)) bw = avl;
      if (bw > 0) bw_ul += bw;
      break;
    default :
       bw = 0;
       break;
  }
  c->lim_bw = bw;
  c->lim_direction = direction;
  return bw;
}

int f4adp_lim_dump(struct child_s* c, char** p, int n)
{ int             i;
  struct limit_s* l;
  char            str[256];

  sprintf(str,"+limit_array=%08X   limit=%08X   count=%i   max=%i   dl=%i   ul=%i",
    (int)limit_array,(int)limit,limit_cnt,limit_max,bw_dl,bw_ul);
  server_reply(c,str);
  for (i=0, l=limit_array; i<limit_cnt; i++, l++)
  { sprintf(str,"+%c %i-%i %02i:%02i-%02i:%02i a=%i,u=%i,s=%i,d=%i,i=%i",
      l == limit ? '*' : ' ',
      l->sday,l->eday,l->shour,l->smin,l->ehour,l->emin,
      l->max_anon,l->max_user,l->max_sum,l->max_anon_dom,l->max_anon_ip);
    server_reply(c,str);
    sprintf(str,"+%c                 u=%i/%i,d=%i/%i,s=%i,us=%i/%i/%i,ds=%i/%i/%i",
      l == limit ? '*' : ' ',
      l->sp_ul,l->sp_ul_fact,l->sp_dl,l->sp_dl_fact,l->sp_sum,
      l->sp_ul_half,l->sp_ul_avg,l->sp_ul_dbl,
      l->sp_dl_half,l->sp_dl_avg,l->sp_dl_dbl);
    server_reply(c,str);
  }
  server_reply(c,"+");
  return 0;
}

int f4adp_lim_getbw(struct child_s* c, char** p, int n)
{ int  i;
  char str[128];

  if (!sscanf(p[0],"%i",&i)) i = -1;
  i = lim_request_bandwidth(c,i);
  sprintf(str,"+%i",i);
  server_reply(c,str);
  return 0;
}

int f4adp_lim_returnbw(struct child_s* c, char** p, int n)
{ lim_return_bandwidth(c);
  server_reply(c,"+ok");
  return 0;
}

/* EOF */
