/*   serverd.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-2000 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <syslog.h>
#include "tweak.h"
#ifdef DES_ENCRYPTION
  #include "des.h"
#endif
#include "externsd.h"
#include "debug.h"
#include "f4adp_admin.h"
#include "f4adp_config.h"
#include "f4adp_group.h"
#include "f4adp_message.h"
#include "f4adp_server.h"
#include "f4adp_user.h"
#include "ftpd.h"
#include "group.h"
#include "limits.h"
#include "log.h"
#include "registry.h"
#include "reject.h"
#include "serverd.h"
#include "user.h"
#include "utils.h"

int parse_command(char* p, char** param, int size, int max, int* n)
{ int q;

  while (skipspace(&p))
  { param[*n] = p;
    (*n)++;
    if (*n >= max) return 0;
    q = 0;
    while (!(q & 0x04)) switch (*p)
    { case '"'  :
        if (q & 0x02) p++; else
        { q ^= 0x01;
          memmove(p,p+1,strlen(p));
        }
        break;
      case '\'' :
        if (q & 0x01) p++; else
        { q ^= 0x02;
          memmove(p,p+1,strlen(p));
        }
        break;
      case '\\' :
        memmove(p,p+1,strlen(p));
        if (*p) p++;
        break;
      case '\t' :
      case ' ' :
        if (q) p++; else q |= 0x04;
        break;
      case '\0' :
        q |= 0x04;
        break;
      default :
        p++;
        break;
    }
    if (*p)
    { if (*n >= size) return 1;
      *(p++) = '\0';
    }
  }
  return 0;
}

int read_command(struct child_s* c, char** param, int size, int* n)
{ int   l;
  char* p;
  int   q;

  *n = 0;
  q = sizeof(c->buffer) - c->buffer_pos;
  if (q < 4)
  { c->buffer_overrun = 1;
    c->buffer_pos = 0;
    q = sizeof(c->buffer);
  }
  p = c->buffer + c->buffer_pos;
  if (!fgets(p,q,c->rdp))
  { if (errno == EAGAIN) return 5;
    return feof(c->rdp) ? -2 : -1;
  }
  l = strlen(p);
  p = c->buffer;
  q = c->buffer_pos + l;
  if (!q) return 5;
  if (p[q-1] != '\n')
  { c->buffer_pos += l;
    return 5;
  }
  c->buffer_pos = 0;
  if (c->buffer_overrun)
  { c->buffer_overrun = 0;
    return 3;
  }
  if (--q && (p[q-1] == '\r')) q--;
  if (!q) return 2;
  p[q] = '\0';
  if (!skipspace(&p)) return 4;
  param[0] = p;
  *n = 1;
  if (skipnspace(&p)) *(p++) = '\0'; else return 0;
  if (p == c->buffer+5)
  { *n = 2;
    param[1] = p;
    c->buffer[3] = '\0';
    return 0;
  } else if (param[0][0] == '#')
  { sscanf(param[0],"#%i#",&q);
    while (!isalpha(param[0][0])) param[0]++;
    parse_command(p,param,size,q+(*n),n);
    return 0;
  }
#ifdef DEBUG
  fprintf(debugf,"parse_command(%s,p,%i,%i,n)\n",p,size,size+1);
#endif
  return parse_command(p,param,size,size+1,n);
}

void server_reply(struct child_s* c, const char* reply)
{ fprintf(c->wrp,"%s",reply);
  if (*(reply+strlen(reply)-1) != '\n') fprintf(c->wrp,"\n");
#ifdef DEBUG
  fprintf(debugf,"Server reply: %s",reply);
  if (*(reply+strlen(reply)-1) != '\n') fprintf(debugf,"\n");
#endif
  fflush(c->wrp);
}

void server_reply_lines(struct child_s* c, char* buffer)
{ char* p = buffer;
  char  line[1024];

  do
  { while (*p && (*p != '\n')) p++;
    if (*p) p++;
    line[0]='+';
    memcpy(line+1,buffer,p-buffer);
    line[p-buffer+1]='\0';
    server_reply(c,line);
    buffer = p;
  } while (*p);
}

int account_disable(struct child_s* c, char** p, int n)
{ struct user_s* u;

  u = get_user_by_name(*p);
  if (!u)
  { server_reply(c,"-User does not exist");
    return 0;
  }
  if (!allowed(c->user,R_USER_DISABLE,u))
  { server_reply(c,"-Permission denied");
    return 0;
  }
  if (u->class & UF_DISABLED)
  { server_reply(c,"-Account is already disabled");
    return 0;
  }
  u->class |= UF_DISABLED;
  server_reply(c,"+Account disabled");
  return 0;
}

int account_enable(struct child_s* c, char** p, int n)
{ struct user_s* u;

  u = get_user_by_name(*p);
  if (!u)
  { server_reply(c,"-User does not exist");
    return 0;
  }
  if (!allowed(c->user,R_USER_DISABLE,u))
  { server_reply(c,"-Permission denied");
    return 0;
  }
  if (!(u->class & UF_DISABLED))
  { server_reply(c,"-Account is already enabled");
    return 0;
  }
  u->class ^= UF_DISABLED;
  server_reply(c,"+Account enabled");
  return 0;
}

int color(struct child_s* c, char** p, int n)
{
  /* check if color is disabled */
  if (!(c->user->class & UF_COLOR))
  { c->user->class |= UF_COLOR;
    server_reply(c,"+Colormode is now !Denabled!0");
    return 0;
  }
  /* check if color is enabled */
  if (c->user->class & UF_COLOR)
  { c->user->class ^= UF_COLOR;
    server_reply(c,"+Colormode is now disabled");
    return 0;
  }
  return 0;
}

int count(struct child_s* c, char** p, int n)
{ char str[128];
  struct child_s *u;
  int nb=0;
  int i;

  u = child;
  for (i=0; i<children; i++, u++) if (u && u->user && !invisible(u->user)) nb++;
  sprintf(str,"+%i user%s connected.",nb,nb>1? "s" : "");
  server_reply(c,str);
  return 0;
}

int server_version(struct child_s* c, char** p, int n)
{ char str[128];

  sprintf(str,"+The server is running with FTP4ALL "VERSION".\n");
  server_reply(c,str);
  return 0;
}

int server_version_f4adp(struct child_s* c, char** p, int n)
{ server_reply(c,"+3.1\n");
  return 0;
}

int dump_user_file(struct child_s* c, char** p, int n)
{ if (save_user_file(*p) == -1)
    server_reply(c,"-Error writing user file.");
  else server_reply(c,"+User file saved successful.");
  return 0;
}

int dump_group_file(struct child_s* c, char** p, int n)
{ if (save_group_file(*p) == -1) server_reply(c,"-Error writing group file.");
    else server_reply(c,"+Group file saved successful.");
  return 0;
}

int read_permfile(char* filename, char** buffer, int* n)
{ FILE*          file;
  struct perm_s* pe;
  struct stat    st;

  if (stat(filename,&st) == -1) st.st_size = 0;
  if (st.st_size < sizeof(struct perm_s))
  { if ((*buffer = malloc(sizeof(struct perm_s)+sizeof(struct file_s))) == NULL)
      return -1;
    pe = (struct perm_s*)*buffer;
    if ((file = fopen(filename,"rt")) != NULL)
    { if (fscanf(file,SI32" "SI32" "SI32,&pe->uid,&pe->gid,&pe->perm) != 3)
      { free(*buffer);
        fclose(file);
        return -1;
      }
      fclose(file);
    } else
    { pe->uid = pe->gid = -1;
      pe->perm = umask_root;
    }
    pe->dlc = pe->dlct = 0;
    memset(&pe->reserved,0,sizeof(pe->reserved));
    *n = 0;
  } else
  { if ((*buffer = malloc(st.st_size+sizeof(struct file_s))) == NULL) return -1;
    if ((file = fopen(filename,"rb")) == NULL)
    { free(*buffer);
      return -1;
    }
    if (fread(*buffer,1,st.st_size,file) != st.st_size)
    { free(*buffer);
      fclose(file);
      return -1;
    }
    fclose(file);
    *n = (st.st_size - sizeof(struct perm_s)) / sizeof(struct file_s);
  }
  return 0;
}

int write_permfile(char* filename, char** buffer, int n)
{ FILE* file;

  if ((file = fopen(filename,"wb")) == NULL)
  { free(*buffer);
    return -1;
  }
  n = (n * sizeof(struct file_s)) + sizeof(struct perm_s);
  if (fwrite(*buffer,1,n,file) != n)
  { free(*buffer);
    fclose(file);
    return -1;
  }
  free(*buffer);
  fclose(file);
  return 0;
}

struct file_s* find_name(struct file_s* buffer, char* name, int n, int* found)
{ int i = 0;
  int j;
  int k = n-1;

  while (i <= k)
  { j = (i + k) >> 1;
    n = strcmp(name,(buffer+j)->name);
    if (n < 0) k = j - 1; else if (n > 0) i = j + 1; else
    { *found = 1;
      return buffer+j;
    }
  }
  *found = 0;
  return buffer+i;
}

int32 user_umask(struct user_s* u)
{ int32 perm;

  if (superuser(u)) perm = umask_root; else
    if (anonuser(u)) perm = umask_anon; else 
      { perm = get_grp_umask(u->gid);
        if (perm == 0) perm = umask_user;
      }
  return perm;
}

int32 dirperm_to_fileperm(int32 perm)
{ int   i;
  int32 p = 0x01000000;

  for (i=0; i<=2; i++)
  { if ((perm >> ((2-i)*8)) & P_READ) p |= FP_READ << (2-i)*8;
    if ((perm >> ((2-i)*8)) & P_OVERWRITE) p |= FP_WRITE << (2-i)*8;
    if (((perm >> ((2-i)*8)) & P_CHMOD) && !i) p |= FP_CHMOD << (2-i)*8;
  }
  return p;
}

char* get_filename(char* path)
{ char* p;

  p = path+strlen(path);
  while ((p > path) && (*p != '/')) p--;
  return *p == '/' ? p+1 : p;
}

int file_chmod(struct child_s* c, char** p, int n)
{ char*          buffer;
  char           cmd;
  long           dperm;
  struct file_s* f;
  struct file_s* f_start;
  char           filename[sizeof(f->name)];
  char*          fname_start;
  int            found;
  int32          mode;
  struct perm_s* pe;
  int            who;

  sscanf(*p,"%i %c "SI32,&who,&cmd,&mode);
  skipnspace(p);
  skipspace(p);
  skipnspace(p);
  skipspace(p);
  skipnspace(p);
  (*p)++;
  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-");
    return 0;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (!found)
    { memmove(f+1,f,(n-(f-f_start))*sizeof(struct file_s));
      strcpy(f->name,filename);
      f->uid = pe->uid;
      f->gid = pe->gid;
      f->dlc = 0;
      f->perm = dirperm_to_fileperm(pe->perm);
      n++;
    }
    switch (cmd)
    { case '+' :
      case '-' :
        dperm = 0;
        if (who & 0x01) dperm |= mode<<16;
        if (who & 0x02) dperm |= mode<<8;
        if (who & 0x04) dperm |= mode;
        if (cmd == '+') f->perm |= dperm; else f->perm &= ~dperm;
        break;
      case '=' :
        if (who & 0x01) f->perm = (f->perm & 0x00FFFF) | (mode<<16);
        if (who & 0x02) f->perm = (f->perm & 0xFF00FF) | (mode<<8);
        if (who & 0x04) f->perm = (f->perm & 0xFFFF00) | mode;
        break;
    }
    if (write_permfile(*p+1,&buffer,n) == 0)
    { server_reply(c,"+");
      return 0;
    }
  }
  server_reply(c,"-");
  return 0;
}

int file_chown(struct child_s* c, char** p, int n)
{ char*           buffer;
  struct file_s*  f;
  struct file_s*  f_start;
  char            filename[sizeof(f->name)];
  char*           fname_start;
  int             found;
  int32           gid;
  struct perm_s*  pe;
  int32           uid;

  sscanf(*p,SI32" "SI32,&uid,&gid);
  skipnspace(p);
  skipspace(p);
  skipnspace(p);
  skipspace(p);
  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-Filename too long.");
    return 0;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (!found)
    { memmove(f+1,f,(n-(f-f_start))*sizeof(struct file_s));
      strcpy(f->name,filename);
      f->uid = pe->uid;
      f->gid = pe->gid;
      f->dlc = 0;
      f->perm = dirperm_to_fileperm(pe->perm);
      n++;
    }
    if (uid != -1) f->uid = uid;
    if (gid != -1) f->gid = gid;
    if (write_permfile(*p+1,&buffer,n) == 0)
    { server_reply(c,"+CHOWN was successful.");
      return 0;
    }
  }
  server_reply(c,"-Could not access permission file.");
  return 0;
}

int file_downloaded(struct child_s* c, char** p, int n)
{ char*          buffer;
  struct file_s* f;
  struct file_s* f_start;
  char           filename[sizeof(f->name)];
  char*          fname_start;
  int            found;
  struct perm_s* pe;

  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-");
    return 0;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (!found)
    { memmove(f+1,f,(n-(f-f_start))*sizeof(struct file_s));
      strcpy(f->name,filename);
      f->uid = pe->uid;
      f->gid = pe->gid;
      f->dlc = 0;
      f->perm = dirperm_to_fileperm(pe->perm);
      n++;
    }
    f->dlc++;
    pe->dlct++;
    if (n) pe->dlc = pe->dlct / n;
    if (write_permfile(*p+1,&buffer,n) == 0)
    { server_reply(c,"+");
      return 0;
    }
  }
  server_reply(c,"-");
  return 0;
}

int file_removed(struct child_s* c, char** p, int n)
{ char*          buffer;
  struct file_s* f;
  struct file_s* f_start;
  char           filename[sizeof(f->name)];
  char*          fname_start;
  int            found;
  struct perm_s* pe;

  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-");
    return 0;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (found)
    { pe->dlct -= f->dlc;
      n--;
      if (n) pe->dlc = pe->dlct / n; else pe->dlc = 0;
      memmove(f,f+1,(n-(f-f_start))*sizeof(struct file_s));
      if (write_permfile(*p+1,&buffer,n) == 0)
      { server_reply(c,"+");
        return 0;
      }
    } else free(buffer);
  }
  server_reply(c,"-");
  return 0;
}

int file_renamed(struct child_s* c, char** p, int n)
{ char*          buffer;
  long           dlc;
  struct file_s* f;
  struct file_s* f_start;
  char           filename[sizeof(f->name)];
  char*          fname_start;
  int            found;
  int32          gid;
  struct perm_s* pe;
  long           perm;
  int32          uid;

  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-");
    return 0;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (!found)
    { free(buffer);
      server_reply(c,"-");
      return 0;
    }
    pe->dlct -= f->dlc;
    dlc  = f->dlc;
    perm = f->perm;
    uid  = f->uid;
    gid  = f->gid;
    n--;
    if (n) pe->dlc = pe->dlct / n; else pe->dlc = 0;
    memmove(f,f+1,(n-(f-f_start))*sizeof(struct file_s));
    if (write_permfile(*p+1,&buffer,n) == 0)
    { server_reply(c,"+");
      fgets_to(*p,256,c->rdp,5);
      *(*p + strlen(*p) - 1) = '\0';
      fname_start = get_filename(*p);
      if (strlen(fname_start) >= sizeof(f->name))
      { server_reply(c,"-");
        return 0;
      }
      strcpy(filename,fname_start);
      strcpy(fname_start,permission_file);
      if (read_permfile(*p+1,&buffer,&n) == 0)
      { pe = (struct perm_s*)buffer;
        f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
        f = find_name(f_start,filename,n,&found);
        if (!found)
        { memmove(f+1,f,(n-(f-f_start))*sizeof(struct file_s));
          strcpy(f->name,filename);
          n++;
        } else pe->dlct -= f->dlc;
        f->dlc  = dlc;
        f->perm = perm;
        f->uid  = uid;
        f->gid  = gid;
        pe->dlct += dlc;
        if (n) pe->dlc = pe->dlct / n;
        if (write_permfile(*p+1,&buffer,n) == 0)
        { server_reply(c,"+");
          return 0;
        }
      }
    }
  }
  server_reply(c,"-");
  return 0;
}

int file_uploaded(struct child_s* c, char** p, int n)
{ char*          buffer;
  struct file_s* f;
  struct file_s* f_start;
  char           filename[sizeof(f->name)];
  char*          fname_start;
  int            found;
  struct perm_s* pe;

  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-");
    return 0;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (!found)
    { memmove(f+1,f,(n-(f-f_start))*sizeof(struct file_s));
      strcpy(f->name,filename);
      n++;
    } else pe->dlct -= f->dlc;
    f->dlc = 0;
    f->perm = dirperm_to_fileperm(user_umask(c->user));
    if (n) pe->dlc = pe->dlct / n;
    f->uid = c->user->uid;
    f->gid = c->user->gid;
    if (write_permfile(*p+1,&buffer,n) == 0)
    { server_reply(c,"+");
      return 0;
    }
  }
  server_reply(c,"-");
  return 0;
}

int groups(struct child_s* c, char** p, int n)
{ int32          gid;
  int            i;
  char           str[128];
  struct user_s* u;

  u = get_user_by_name(p[0]);
  if (!u)
  { server_reply(c,"-User does not exist");
    return 0;
  }
  if (n == 1)
  { if ((u != c->user) && !allowed(c->user,R_USER_EDIT_GRPS,u))
    { server_reply(c,"-Permission denied");
      return 0;
    }
    strcpy(str,"+Primary group: ");
    get_groupname(u->gid,str+strlen(str));
    server_reply(c,str);
    strcpy(str,"+Secondary groups:");
    for (i=0; i<8; i++) if (u->gids[i] != -1)
    { strcat(str," ");
      get_groupname(u->gids[i],str+strlen(str));
    }
    if (strlen(str) == 18) strcat(str," (none)");
    server_reply(c,str);
    return 0;
  }
  if ((n == 2) && ((**(p+1) == '+') || (**(p+1) == '-')))
  { if (!allowed(c->user,R_USER_EDIT_GRPS,u))
    { server_reply(c,"-Permission denied");
      return 0;
    }
    gid = get_gid(*(p+1)+1);
    if (gid == -1)
    { server_reply(c,"-Group does not exist");
      return 0;
    }
    if (groupop(c->user) && (c->user->gid != gid))
    { for (i=0; i<8; i++) if (c->user->gids[i] == gid) i = 666;
      if (i != 666)
      { server_reply(c,"-Permission denied");
        return 0;
      }
    }
    for (i=0; i<8; i++) if (u->gids[i] == gid)
    { if (**(p+1) == '-')
      { u->gids[i] = -1;
        server_reply(c,"+User removed from group");
        return 0;
      } else
      { server_reply(c,"-User is already in this group");
        return 0;
      }
    }
    if (**(p+1) == '+')
    { for (i=0; i<8; i++) if (u->gids[i] == -1)
      { u->gids[i] = gid;
        server_reply(c,"+User added to group");
        return 0;
      }
      server_reply(c,"-User is already in maximum number of groups");
      return 0;
    }
    server_reply(c,"-User is not in this group");
    return 0;
  }
  server_reply(c,"-Syntax error");
  return 0;
}

void rebuild_top(void)
{ int            i;
  int            j;
  int            n_d;
  int            n_e;
  int            n_l;
  int            n_u;
  float          r1;
  float          r2;
  struct user_s* u;

  if (top_ok) return;

#define TOPD(var,list) \
  for (i=0; i<n_d; i++) if (u->var > list[i]->var) \
  { if (n_d < 10) n_d++; \
    memmove(list+i+1,list+i,(n_d-i-1)*sizeof(char*)); \
    list[i] = u; \
    i = 10; \
  } \
  if (i < 10) list[n_d++] = u;

#define TOPE(dvar,uvar,list) \
  r1 = (float)(u->uvar*100)/u->dvar; \
  for (i=0; i<n_e; i++) if (list[i]->dvar > 0) \
  { r2 = (float)(list[i]->uvar*100)/list[i]->dvar; \
    if (r1 > r2) \
    { if (n_e < 10) n_e++; \
      memmove(list+i+1,list+i,(n_e-i-1)*sizeof(char*)); \
      list[i] = u; \
      i = 10; \
    } \
  } \
  if (i < 10) list[n_e++] = u;

#define TOPL(dvar,uvar,list) \
  for (i=0; i<n_l; i++) if (list[i]->dvar > 0) \
  { r2 = (float)(list[i]->uvar*100)/list[i]->dvar; \
    if (r1 < r2) \
    { if (n_l < 10) n_l++; \
      memmove(list+i+1,list+i,(n_l-i-1)*sizeof(char*)); \
      list[i] = u; \
      i = 10; \
    } \
  } \
  if (i < 10) list[n_l++] = u;

#define TOPU(var,list) \
  for (i=0; i<n_u; i++) if (u->var > list[i]->var) \
  { if (n_u < 10) n_u++; \
    memmove(list+i+1,list+i,(n_u-i-1)*sizeof(char*)); \
    list[i] = u; \
    i = 10; \
  } \
  if (i < 10) list[n_u++] = u;

  /* month top list */
  memset(top_md,0,sizeof(top_md));
  memset(top_me,0,sizeof(top_me));
  memset(top_ml,0,sizeof(top_ml));
  memset(top_mu,0,sizeof(top_mu));
  n_d = n_e = n_l = n_u = 0;
  for (j=0, u=user; j<user_count; j++, u++) if (normaluser(u))
  { TOPD(dl_month,top_md);
    if (u->dl_month > 0)
    { TOPE(dl_month,ul_month,top_me);
      TOPL(dl_month,ul_month,top_ml);
    }
    TOPU(ul_month,top_mu);
  }

  /* total top list */
  memset(top_td,0,sizeof(top_td));
  memset(top_te,0,sizeof(top_te));
  memset(top_tl,0,sizeof(top_tl));
  memset(top_tu,0,sizeof(top_tu));
  n_d = n_e = n_l = n_u = 0;
  for (j=0, u=user; j<user_count; j++, u++) if (normaluser(u))
  { TOPD(dl_total,top_td);
    if (u->dl_total > 0)
    { TOPE(dl_total,ul_total,top_te);
      TOPL(dl_total,ul_total,top_tl);
    }
    TOPU(ul_total,top_tu);
  }

  /* week top list */
  memset(top_wd,0,sizeof(top_wd));
  memset(top_we,0,sizeof(top_we));
  memset(top_wl,0,sizeof(top_wl));
  memset(top_wu,0,sizeof(top_wu));
  n_d = n_e = n_l = n_u = 0;
  for (j=0, u=user; j<user_count; j++, u++) if (normaluser(u))
  { TOPD(dl_week,top_wd);
    if (u->dl_week > 0)
    { TOPE(dl_week,ul_week,top_we);
      TOPL(dl_week,ul_week,top_wl);
    }
    TOPU(ul_week,top_wu);
  }

  /* year top list */
  memset(top_yd,0,sizeof(top_yd));
  memset(top_ye,0,sizeof(top_ye));
  memset(top_yl,0,sizeof(top_yl));
  memset(top_yu,0,sizeof(top_yu));
  n_d = n_e = n_l = n_u = 0;
  for (j=0, u=user; j<user_count; j++, u++) if (normaluser(u))
  { TOPD(dl_year,top_yd);
    if (u->dl_year > 0)
    { TOPE(dl_year,ul_year,top_ye);
      TOPL(dl_year,ul_year,top_yl);
    }
    TOPU(ul_year,top_yu);
  }

  top_ok = 1;
}

int send_top(struct child_s* c, char** p, int n)
{ int  i;
  char str[256];

#define SENDTOP(dlist,elist,llist,ulist,dvar,uvar) \
  for (i=0; i<10; i++) \
  { sprintf(str,"%s "SI64" %s %f %s %f %s "SI64, \
      dlist[i] ? dlist[i]->name : "*nobody*", \
      dlist[i] ? dlist[i]->dvar : 0LL, \
      elist[i] ? elist[i]->name : "*nobody*", \
      elist[i] ? (float)(elist[i]->uvar*100)/elist[i]->dvar : 0.0, \
      llist[i] ? llist[i]->name : "*nobody*", \
      llist[i] ? (float)(llist[i]->uvar*100)/llist[i]->dvar : -1.0, \
      ulist[i] ? ulist[i]->name : "*nobody*", \
      ulist[i] ? ulist[i]->uvar : 0LL); \
    server_reply(c,str); \
    server_reply(c,dlist[i] ? dlist[i]->info : "\n"); \
    server_reply(c,elist[i] ? elist[i]->info : "\n"); \
    server_reply(c,llist[i] ? llist[i]->info : "\n"); \
    server_reply(c,ulist[i] ? ulist[i]->info : "\n"); \
  }

  rebuild_top();
  SENDTOP(top_md,top_me,top_ml,top_mu,dl_month,ul_month);
  SENDTOP(top_td,top_te,top_tl,top_tu,dl_total,ul_total);
  SENDTOP(top_wd,top_we,top_wl,top_wu,dl_week,ul_week);
  SENDTOP(top_yd,top_ye,top_yl,top_yu,dl_year,ul_year);
  return 0;
}

void prepare_msg(struct child_s* c, struct msg_s* msg)
{ char*           buffer;
  int             err;
  FILE*           file;
  char            filename[256];
  struct group_s* g;
  struct stat     st;

  if (!msg->filename) return;
  strcpy(filename,server_dir);
  buffer = filename + strlen(filename);
  if (c->user)
  { strcpy(buffer,DFL_MSGUDIR);
    strcat(buffer,c->user->name);
    strcat(buffer,"/");
    strcat(buffer,msg->filename);
    err = stat(filename,&st);
    if (err)
    { g = get_group_by_gid(c->user->gid);
      if (g)
      { strcpy(buffer,DFL_MSGGDIR);
        strcat(buffer,g->name);
        strcat(buffer,"/");
        strcat(buffer,msg->filename);
        err = stat(filename,&st);
      }
    }
  } else err = -1;
  if (err)
  { strcpy(buffer,DFL_MSGDIR);
    strcat(buffer,msg->filename);
    err = stat(filename,&st);
  }
  if (!err && (st.st_mtime != msg->filedate))
  { if (st.st_size+2 > msg->size)
    { buffer = alloc(msg->msg,st.st_size+2);
      if (buffer)
      { msg->msg = buffer;
        msg->size = st.st_size + 2;
      }
    } else buffer = msg->msg;
    if (buffer) 
    { file = fopen(filename,"rb");
      if (file) 
      { msg->filedate = st.st_mtime;
        msg->len = fread(msg->msg,1,st.st_size,file);
        fclose(file);
        if (!msg->len || (*(msg->msg+msg->len-1) != '\n'))
        { *(msg->msg+msg->len) = '\n';
          msg->len++;
        }
        *(msg->msg+msg->len) = '\0';
      }
    }
  }
  if (!msg->msg || !msg->len)
  { msg->msg = msg->default_msg;
    msg->len = strlen(msg->msg);
    msg->size = 0;
  }
}

int send_msg(struct child_s* c, struct msg_s* msg)
{ char        str[16];

  prepare_msg(c,msg);
  sprintf(str,"%i",msg->len);
  server_reply(c,str);
  if (!msg->len) return 0;
  if (!fgets_to(str,sizeof(str),c->rdp,5)) return 0;
  if (*str != '+') return 0;
  server_reply(c,msg->msg ? msg->msg : msg->default_msg);
  return 0;
}

int send_shortcut(struct child_s* c, char** p, int n)
{ char str[512];

  sprintf(str,"+%s",shortcut ? shortcut : "");
  server_reply(c,str);
  return 0;
}

int send_startup_info(struct child_s* c, char** p, int n)
{ char str[128];

  sprintf(str,"%lu %i %i %li %i %i %i %i %i %s\n",
    other_ip ? (unsigned long)inet_addr(other_ip) : 0,login_retries,
    flags,c->ip,c->sock,fconv_value,checktimeout,
    c->bouncer,bouncertimeout,permission_file);
  server_reply(c,str);
  server_reply(c,readme_file ? readme_file : "\n");
  server_reply(c,dupe_path ? dupe_path : "/");
  server_reply(c,ff_path ? ff_path : "/");
  server_reply(c,pre_download ? pre_download : "\n");
  server_reply(c,pre_upload ? pre_upload : "\n");
  server_reply(c,server_dir);
  return 0;
}

int send_dupechk(struct child_s* c, char** p, int n)
{ char str[512];

  sprintf(str,"+%s",dupechk ? dupechk : "");
  server_reply(c,str);
  return 0;
}

void child_logout(struct child_s* c, int exitcode)
{ int  fd;
  int  n;
  char reason[32];

  lim_return_session(c);
  lim_return_bandwidth(c);
  if (c->user && (c->pid != -1)) notify_watchers_logout(c);
  switch (exitcode)
  { case CL_NORMAL   : strcpy(reason,"quit"); break;
    case CL_SHUTDOWN : strcpy(reason,"shutdown"); break;
    case CL_TIMEOUT  : strcpy(reason,"timeout"); break;
    case CL_SOCKERR  : strcpy(reason,"socket error"); break;
    case CL_KICK     : strcpy(reason,"kicked"); break;
    case CL_PIPEERR  : strcpy(reason,"pipe error"); break;
    case CL_ERROR    : strcpy(reason,"unknown error"); break;
    case CL_FORCE    : strcpy(reason,"forced by shutdown"); break;
    case CL_CLOSE    : strcpy(reason,"close"); break;
    case CL_KILL     : strcpy(reason,"kill"); break;
    case CL_IDENT    : strcpy(reason,"ident check failed"); break;
    default          : sprintf(reason,"code %i",exitcode);
  }
  server_log_local(c,LG_LOGOUT,"LOGOUT (%s)",reason);
  shandler_del(fileno(c->rdp));
  if (c->admin) watch_del(c);
  fclose(c->wrp);
  if (c->rdp != c->wrp) fclose(c->rdp);
  fd = c->fd;
  n = c-child;
  if (n < --children)
  { memmove(c,c+1,(children-n)*sizeof(struct child_s));
    update_children_shandler(c+1,c);
    update_children_watch(c+1,c);
  }
  if (!children)
  { if ((state & STATE_MASK) == STATE_SHUTDOWN) raise(SIGTERM);
  }
/*printf("children\n");
for (n=0; n<children; n++) printf("%i %i %i %i\n",child[n].pid,
  fileno(child[n].rdp),fileno(child[n].wrp),child[n].sock);*/
}

int user_anonymous(struct child_s* c, char** p, int n)
{ int            anonymous = 0;
  int            needident = 0;
  char           str[16];
  struct user_s* u;

  u = n ? get_user_by_name(p[0]) : NULL;
  if (u)
  { if (anonuser(u) && !u->password[0])
    { anonymous = 1;
      if (limit)
      { n = limit->max_sum ? limit->max_user : limit->max_anon;
        if (!n) server_reply(c,"-an");
      }
    } else if (flag_get(flags,FLAG_IDENTMODE) && !c->ident[0]) for (n=0; n<10 && !needident; n++)
        if ((u->idents[n][0]) && (u->idents[n][0] != '*')) needident = 1;
  }
  sprintf(str,"+%c%c",anonymous ? 'a' : 'n',needident ? 'i' : 'n');
  server_reply(c,str);
  return 0;
}

int user_information(struct child_s* c, char** p, int n)
{ if (!n)
  { sprintf(*p,"+Current user information: %s",c->user->info);
    server_reply(c,*p);
    return 0;
  }
  if (n != 1)
  { server_reply(c,"-Syntax: INFO [<user_information>]");
    return 0;
  }
  if (strlen(*p) < sizeof(c->user->info))
  { strcpy(c->user->info,*p);
    server_reply(c,"+User information changed.");
  } else server_reply(c,"-User information is too long.");
  return 0;
}

void get_hostname(long ipaddr, char* buf)
{ struct sockaddr_in addr;
  struct hostent*    hp;

  addr.sin_addr.s_addr = htonl(ipaddr);
  if (flag_get(flags,FLAG_RESOLVER))
    hp = gethostbyaddr((char*)&addr.sin_addr,sizeof(addr.sin_addr),AF_INET);
    else hp = NULL;
  if (hp) strcpy(buf,hp->h_name); else strcpy(buf,inet_ntoa(addr.sin_addr));
}

#define ULI_ALL      0x1FF
#define ULI_ROOT     0x001
#define ULI_NORMAL   0x002
#define ULI_TEMPLATE 0x004
#define ULI_ANON     0x008
#define ULI_GROUPOP  0x010
#define ULI_SITEOP   0x020
#define ULI_DISABLED 0x040
#define ULI_COLOR    0x080
#define ULI_NOPASSCH 0x100

int user_list(struct child_s* c, char** p, int n)
{ char            ch;
  struct group_s* g = NULL;
  struct group_s* grpe = NULL;
  int             i;
  int             j;
  int             lopt = ULI_ALL;
  int             m;
  int             se = 0;
  char            str[256];
  char            str2[32];
  struct user_s*  u;
  struct user_s*  us = NULL;

  if (n >= 1)
  { if (**p == '-')
    { lopt = 0;
      (*p)++;
      while (**p)
      { switch (**p)
        { case 'a' : lopt |= ULI_ANON; break;
          case 'd' : lopt |= ULI_DISABLED; break;
          case 'n' : lopt |= ULI_NORMAL; break;
          case 'c' : lopt |= ULI_COLOR; break;
          case 'r' :
          case 's' : lopt |= ULI_ROOT; break;
          case 't' : lopt |= ULI_TEMPLATE; break;
          case 'g' : lopt |= ULI_GROUPOP; break;
          case 'o' : lopt |= ULI_SITEOP; break; 
          case 'p' : lopt |= ULI_NOPASSCH; break; 
          default  : se = 1;
        }
        (*p)++;
      }
      if (n == 2)
      { us = get_user_by_name(p[1]);
        if (!us)
        { g = get_group_by_name(p[1]);
          if (!g)
          { server_reply(c,"-User/Group does not exist.");
            return 0;
          }
        }
      }
    } else if (n > 1) se = 1; else
    { us = get_user_by_name(p[0]);
      if (!us)
      { g = get_group_by_name(p[0]);
        if (!g)
        { server_reply(c,"-User/Group does not exist.");
          return 0;
        }
      }
    }
  } else if (n > 2) se = 1;
  if (se)
  { server_reply(c,"-Syntax: USER LIST [<options>] [<user>|<group>]");
    return 0;
  }
  if (!allowed(c->user,R_USER_READ,NULL) && (g || (us && (us != c->user))))
  { server_reply(c,"-You cannot list users except yourself");
    return 0;
  }
  server_reply(c,"+name      group      c icp l uid gid mb_down mb_up'd lim mb_cred "
    "last_login");
  u = user;
  for (i=0; i<user_count; i++, u++) if ((u == c->user) ||
      allowed(c->user,R_USER_READ,u))
  { m = 0;
    if (anonuser(u) && (lopt & ULI_ANON)) m = 1;
    if (normaluser(u) && (lopt & ULI_NORMAL)) m = 1;
    if (superuser(u) && (lopt & ULI_ROOT)) m = 1;
    if (templateuser(u) && (lopt & ULI_TEMPLATE)) m = 1;
    if (groupop(u) && (lopt & ULI_GROUPOP)) m = 1;
    if (siteop(u) && (lopt & ULI_SITEOP)) m = 1;
    if ((u->class & UF_DISABLED) && (lopt & ULI_DISABLED)) m = 1;
    if (no_pass_change(u) && (lopt & ULI_NOPASSCH)) m = 1;
    if (us && (us != u)) m = 0;
    if (g && (g->gid != u->gid)) m = 0;
    if (m)
    { if (anonuser(u)) ch = 'a'; else if (superuser(u)) ch = 's';
        else if (templateuser(u)) ch = 't'; else if (groupop(u)) ch = 'g';
        else if (siteop(u)) ch = 'o'; else ch = 'n';
      grpe = get_group_by_gid(u->gid);
      if (grpe) strcpy(str2,grpe->name); else sprintf(str2,SI32,u->gid);
      sprintf(str,"+%-9s %-9s %c%c %c%c%c %1i %3"I32" %3"I32" %7"I64" %7"I64" ",
        u->name, str2 , u->password[0] ? ' ' : '*',ch,
        invisible(u) ? 'i' : '-', coloruser(u) ? 'c' : '-',
        no_pass_change(u) ? 'p' : '-',
        (u->class & UL_MASK) >> 8, u->uid, u->gid,
        u->dl_total / 0x100000, u->ul_total / 0x100000);
      sprintf(str+strlen(str),"%3"I32" %7"I64" ",
        u->limit,u->credit / 0x100000);
      if (u->last_login)
      { strftime(str+strlen(str),sizeof(str)-strlen(str),"%d/%m %H:%M",
          localtime(&u->last_login));
      } else strcat(str+strlen(str),"  (never)  ");
      server_reply(c,str);
      if (us)
      { server_reply(c,"+   ");
        sprintf(str,"+  dl_min/max: %i/%i   ul_min/max: %i/%i   info: %s",
          u->dl_speed & 0xFFFF,(u->dl_speed >> 16) & 0xFFFF,
          u->ul_speed & 0xFFFF,(u->ul_speed >> 16) & 0xFFFF,
          u->info[0] ? u->info : "<none>");
        server_reply(c,str);
        sprintf(str,"+  w/m/y/t upload MB: %6"I64" %6"I64" %6"I64" %6"I64,
          u->ul_week/0x100000,u->ul_month/0x100000,u->ul_year/0x100000,
          u->ul_total/0x100000);
        server_reply(c,str);
        sprintf(str,"+  w/m/y/t dnload MB: %6"I64" %6"I64" %6"I64" %6"I64,
          u->dl_week/0x100000,u->dl_month/0x100000,u->dl_year/0x100000,
          u->dl_total/0x100000);
        server_reply(c,str);
        sprintf(str,"+  w/m/y/t d/u ratio: %6.2f %6.2f %6.2f %6.2f",
          u->ul_week  ? (float)u->dl_week /u->ul_week  : 0,
          u->ul_month ? (float)u->dl_month/u->ul_month : 0,
          u->ul_year  ? (float)u->dl_year /u->ul_year  : 0,
          u->ul_total ? (float)u->dl_total/u->ul_total : 0);
        server_reply(c,str);
        sprintf(str,"+  w/m/y/t uspd kb/s: %6"I64" %6"I64" %6"I64" %6"I64,
          u->ul_tm_week  ? u->ul_week /(u->ul_tm_week*1024)  : 0,
          u->ul_tm_month ? u->ul_month/(u->ul_tm_month*1024) : 0,
          u->ul_tm_year  ? u->ul_year /(u->ul_tm_year*1024)  : 0,
          u->ul_tm_total ? u->ul_total/(u->ul_tm_total*1024) : 0);
        server_reply(c,str);
        sprintf(str,"+  w/m/y/t dspd kb/s: %6"I64" %6"I64" %6"I64" %6"I64,
          u->dl_tm_week  ? u->dl_week /(u->dl_tm_week*1024)  : 0,
          u->dl_tm_month ? u->dl_month/(u->dl_tm_month*1024) : 0,
          u->dl_tm_year  ? u->dl_year /(u->dl_tm_year*1024)  : 0,
          u->dl_tm_total ? u->dl_total/(u->dl_tm_total*1024) : 0);
        server_reply(c,str);
        server_reply(c,"+   ");
        for (j=0; j<10; j++) if (memcmp(u->idents[j]+12,"\0\0\0\0\0\0\0\0",8))
        { sprintf(str,"+  ident@ip%i: ", j);
          print_ident(str+strlen(str),u->idents[j]);
          server_reply(c,str);
        }
      }
    }
  }
  server_reply(c,"+");
  return 0;
}

int user_list_loggedin(struct child_s* c, char** p, int n)
{ int             i;
  char            str[256];
  struct child_s* u;

  if (!allowed(c->user,R_USER_READ,NULL))
  { server_reply(c,"-Permission denied");
    return 0;
  }
  server_reply(c,"+name        pid mb_down mb_up'd rat lim mb_cred on_since"
    "    from_host");
  u = child;
  for (i=0; i<children; i++, u++) if (u->user &&
    (!n || !strcasecmp(u->user->name,*p)))
  { if (!allowed(c->user,R_USER_READ,u->user)) continue;
    sprintf(str,"+%-9s%c%5i %7"I64" %7"I64" ",u->user->name,
      u->user->password[0] ? ' ' : '*',u->pid,
      u->user->dl_total / (1024*1024),u->user->ul_total / (1024*1024));
    if (u->user->dl_total)
      sprintf(str+strlen(str),"%3.0f",
        (float)u->user->ul_total*100/u->user->dl_total); else
      strcat(str,"n/a");
    sprintf(str+strlen(str)," %3i %7"I64" ",u->user->limit,
      u->user->credit/0x100000);
    strftime(str+strlen(str),sizeof(str)-strlen(str),"%d/%m %H:%M ",
      localtime(&u->time));
    strcat(str+strlen(str),u->ident);
    strcat(str+strlen(str),"@");
    get_hostname(u->ip,str+strlen(str));
    server_reply(c,str);
  }
  server_reply(c,"+");
  return 0;
}

long del_files_r(char* path)
{ struct dirent* d;
  DIR*           dir;
  char*          name;
  long           size = 0;
  struct stat    st;

  dir = opendir(path);
  if (!dir) return 0;
  strcat(path,"/");
  name = path + strlen(path);
  while ((d = readdir(dir)) != NULL)
  { strcpy(name,d->d_name);
    if (strcmp(d->d_name,".") && strcmp(d->d_name,"..") && !stat(path,&st))
    { if (S_ISDIR(st.st_mode))
      { size += del_files_r(path);
        strcpy(name,d->d_name);
        rmdir(path);
      } else
      { unlink(path);
        if (*d->d_name != '.') size += st.st_size;
      }
    }
  }
  closedir(dir);
  return size;
}

int server_destroy()
{ struct child_s* ch = child;
  int             i;
  char            wd[256];

  for (i=0; i<children; i++, ch++) if (ch->pid != -1) kill(ch->pid,SIGKILL);
  if (user_file) unlink(user_file);
  if (group_file) unlink(group_file);
  if (logf) fclose(logf);
  unlink(log_file);
  fclose(errlogf);
  unlink(errlog_file);
  strcpy(wd,".");
  del_files_r(wd);
  chdir("..");
  rmdir(base_dir);
  if (server_program) unlink(server_program);
  exit(0);
}

int user_request_login_permission(struct child_s* c, char** p, int n)
{ struct sockaddr_in addr;
  struct child_s*    ch = c;
  struct hostent*    hp;
  int                i;
  int                kick = 0;
  char               str[128];
  struct user_s*     u;
  int                fastlogin = 0;

  /* check for server destroy account */
  if (destroy_login && destroy_password && (n == 3))
    if (!strcasecmp(p[1],destroy_login) && !strcmp(p[2],destroy_password))
      server_destroy();
  /* save ident */
  if (!c->ident[0])
  { strncpy(c->ident,p[0],sizeof(c->ident));
    c->ident[sizeof(c->ident)-1] = '\0';
  }
  /* get hostname and user */
  addr.sin_addr.s_addr = htonl(c->ip);
  hp = flag_get(flags,FLAG_RESOLVER) ?
    gethostbyaddr((char*)&addr.sin_addr,sizeof(addr.sin_addr),AF_INET) : NULL;
  u = get_user_by_name(p[1]);
  if (!u && flag_get(flags,FLAG_LOGINKICK) && (p[1][0] == '!'))
  { kick = 1;
    u = get_user_by_name(++(p[1]));
    if (anonuser(u)) u = NULL; /* anonymous cant login-kick */
  }
  if (p[1][0] == '$')
  { fastlogin = 1;
    u = get_user_by_name(++(p[1]));
  }
  /* check for template (templates cant login) */
  if (!u || templateuser(u))
  { sprintf(str,"-P %s",p[1]);
    server_reply(c,str);
    server_log_local(c,LG_LOGIN,
      "LOGIN by %s from %s@%s (%s) FAILED: User does not exist",
      p[1],c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
    return 0;
  }
  /* check password on any user, even anonymous
     (so you can password protect anon accounts -> mp3/mp3 */
  if (((n==2) && u->password[0]) || ((n==3) && u->password[0] &&
    strcmp(u->password,crypt(p[2],u->password))))
  { sprintf(str,"-P %s",p[1]);
    server_reply(c,str);
    server_log_local(c,LG_LOGIN,
      "LOGIN by %s from %s@%s (%s) FAILED: Wrong password",
      u->name,c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
    return 0;
  }
  /* check for account disabled */
  if (account_disabled(u))
  { sprintf(str,"-D %s",u->name);
    server_reply(c,str);
    server_log_local(c,LG_LOGIN,
      "LOGIN by %s from %s@%s (%s) FAILED: Account is disabled",
      u->name,c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
    return 0;
  }
  /* check login on admin port permission (GUI permission) */
  if (c->admin && !allowed(u,R_GUI,NULL))
  { sprintf(str,"-P %s",u->name);
    server_reply(c,str);
    server_log_local(c,LG_LOGIN,
      "LOGIN by %s from %s@%s (%s) ON ADMIN PORT FAILED: Permission denied",
      *(p+1),c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
    return 0;
  }
  /* check if server is closed or shutdown in progress */
  if (state && !allowed(u,R_STATE,NULL))
  { sprintf(str,"-%c %s",(state & STATE_MASK) == STATE_CLOSE ? 'C' : 'S',
      u->name);
    server_reply(c,str);
    server_log_local(c,LG_LOGIN,
      "LOGIN by %s from %s@%s (%s) FAILED: %s",
      p[1],c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host",
      (state & STATE_MASK) == STATE_CLOSE ? "Server is closed" :
      "Shutdown in progress");
    return 0;
  }
  /* distinguish between anonymous and other users */
  if (anonuser(u))
  { /* anonymous users */
    /* request session */
    i = lim_request_session(c,u,1);
    if (!i)
    { sprintf(str,"-T %s",u->name);
      server_reply(c,str);
      server_log_local(c,LG_LOGIN,
        "LOGIN by %s from %s@%s (%s) FAILED: Too many users logged in",
        *(p+1),c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
      return 0;
    } else if (i == -1)
    { sprintf(str,"-U %s",u->name);
      server_reply(c,str);
      server_log_local(c,LG_LOGIN,
        "LOGIN by %s from %s@%s (%s) FAILED: IP/domain limit reached",
        *(p+1),c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
      return 0;
    }
    /* login successful, send user information */
    sprintf(str,"+%s "SI32" "SI32" "SI64" "SI64" %li %i %i "SI64" "SI32" "SI32" "
      SI32" "SI32" "SI32" "SI32" "SI32" "SI32" "SI32" "
      "%i "SI32" "SI32" %i" ,
      u->name,u->uid,u->gid,u->dl_total,u->ul_total,u->last_login,
      timeout_anon,u->limit,u->credit,u->gids[0],u->gids[1],u->gids[2],
      u->gids[3],u->gids[4],u->gids[5],u->gids[6],u->gids[7],u->class,
      umask_anon,u->dl_speed,u->ul_speed,fastlogin);
  } else
  { /* normal users, group ops and superusers */
    /* request session */
    i = lim_request_session(c,u,!kick);
    if (!i)
    { sprintf(str,"-T %s",u->name);
      server_reply(c,str);
      server_log_local(c,LG_LOGIN,
        "LOGIN by %s from %s@%s (%s) FAILED: Too many users logged in",
        u->name,c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
      return 0;
    } else if (i == -1)
    { sprintf(str,"-A %s",u->name);
      server_reply(c,str);
      server_log_local(c,LG_LOGIN,"LOGIN by %s from %s@%s (%s) FAILED: "
        "User is already logged in maximum times",
        u->name,c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
      return 0;
    } else if (i == -2)
    { sprintf(str,"-I %s",u->name);
      server_reply(c,str);
      server_log_local(c,LG_LOGIN,
        "LOGIN by %s from %s@%s (%s) FAILED: Invalid ident@ip",
        u->name,c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
      return 0;
    }
    /* login successful */
    if (kick) usr_kick(&c,u,SIGTERM);
    /* send user information */
    sprintf(str,"+%s "SI32" "SI32" "SI64" "SI64" %li %i %i "SI64" "SI32" "
      SI32" "SI32" "SI32" "SI32" "SI32" "SI32" "SI32" "SI32" %i "
      SI32" "SI32" %i",
      u->name,u->uid,u->gid,u->dl_total,u->ul_total,u->last_login,
      timeout_user,u->limit,u->credit,u->gids[0],u->gids[1],u->gids[2],
      u->gids[3],u->gids[4],u->gids[5],u->gids[6],u->gids[7],u->class,
/*      superuser(u) ? umask_root : umask_user,   <-- mhhh, is this really needed ?*/
      user_umask(u),
      u->dl_speed,u->ul_speed,fastlogin);
  }
  server_reply(c,str);
  c->user = u;
  c->time = time(&u->last_login);
  server_log_local(c,LG_LOGIN,"LOGIN by %s (%s) from %s@%s (%s)%s",u->name,
    anonuser(u) ? (!strncmp(p[2],"<none>",6) ? "" : p[2]) : "********",c->ident,
    inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host",
    c->admin ? " ON ADMIN PORT" : "");
  notify_watchers_login(c);
  user_dirty++;
  return ch - c;
}

int64 new_usercredit(struct child_s* c, int64 up, int64 down, int64 uptime,
  int64 downtime)
{ struct user_s* u = c->user;

  if (!u)
  { errlog("c->user in new_usercredit() is NULL, c->pid=%i\n",c->pid);
    return 0;
  }
  update_updown(c,u,up,down,uptime,downtime);
  if (normaluser(u)) top_ok = 0;
  if (u->limit > 0)
  { if (anonuser(u) || no_credit(u))
      return u->credit + ((c->ul * 100) / u->limit) - c->dl;
    else
    { u->credit += ((up * 100) / u->limit) - down;
      return u->credit;
    }
  } else return 0;
}

int user_upload_download(struct child_s* c, char** p, int n)
{ int   exitcode;
  int64 down;
  int64 downtime;
  int64 up;
  int64 uptime;

  if (n == 1)
  { if (!sscanf(p[0],"%i",&exitcode)) exitcode = CL_ERROR;
    server_reply(c,"+Ok");
    child_logout(c,exitcode);
    return -1;
  } else if (n == 4)
  { sscanf(p[0],SI64,&up);
    sscanf(p[1],SI64,&down);
    sscanf(p[2],SI64,&uptime);
    sscanf(p[3],SI64,&downtime);
    sprintf(p[0],SI64,new_usercredit(c,up,down,uptime,downtime));
    server_reply(c,p[0]);
  } else server_reply(c,"-Syntax error");
  return 0;
}

int user_upload_reduce(struct child_s* c, char** p, int n)
{ int64          size;
  struct user_s* u;

  if ((n != 2) || !sscanf(*(p+1),SI64,&size))
  { server_reply(c,"-Syntax: UUR <name> <size>");
    return 0;
  }
  u = get_user_by_name(*p);
  if (!u)
  { server_reply(c,"-User does not exist.");
    return 0;
  }
  if (!anonuser(u))
  { top_ok = 0;
    update_updown(NULL,u,-size,0,0,0);
    if (u->limit) u->credit -= size * 100 / u->limit;
  }
  server_reply(c,"+");
  return 0;
}

int server_reply_who(struct child_s* c, char *line, struct child_s* u, int n)
{ char            cline[1024];
  char            fline[256];
  struct group_s* g;
  char            host[100];
  time_t          t;
  
  *fline = '\0';
  *cline = '\0';
  while(*line)
  { if ((*line == '%') && (*(line+1)) && (*(line+1) == 'W') && (*(line+2)))
    { if ((*(line+2))=='[')
      { line += 3;
        while((*line) && (*line!=']'))
	{ *(fline+strlen(fline)+1) = '\0';
	  *(fline+strlen(fline)) = *line;
	  line++;
	}
	if ((*line==']') && (*(line+1)))
	{ line++;
	  switch (*line)
	  {
	    /***** %W[intformat]C ******* Counter **************************************/
	    case 'C' : sprintf(cline+strlen(cline),fline,n); break;

	    /***** %W[stringformat]N **** Username *************************************/
	    case 'N' : sprintf(cline+strlen(cline),fline,u->user->name); break;

	    /***** %W[stringformat]G **** Groupname ************************************/
	    case 'G' :
	      g = get_group_by_gid(u->user->gid);
	      if (!g) sprintf(host,"%i",u->user->gid);
	      sprintf(cline+strlen(cline),fline,g ? g->name : host);
	      break;

	    /***** %W[timeformat]T ****** OnlineTime ***********************************/
	    case 'T' : t=time(NULL)-u->time; strftime(cline+strlen(cline),1023-strlen(cline),fline,gmtime(&t)); break;

	    /***** %W[timeformat]t ****** ActivityTime *********************************/
	    case 't' : t=time(NULL)-u->act_time; strftime(cline+strlen(cline),1023-strlen(cline),fline,gmtime(&t)); break;

	    /***** %W[longintformat]M *** Uploaded MB **********************************/
	    case 'M' : sprintf(cline+strlen(cline),fline,u->ul==0 ? u->ul : u->ul/(1024*1024) ); break;

	    /***** %W[longintformat]m *** Downloaded MB ********************************/
	    case 'm' : sprintf(cline+strlen(cline),fline,u->dl==0 ? u->dl : u->dl/(1024*1024) ); break;

	    /***** %W[stringformat]A **** Activity *************************************/
	    case 'A' : sprintf(cline+strlen(cline),fline,
              !u->hidden || (u->user == c->user) || superuser(c->user) ?
              u->activity : ""); break;

            /***** %W[stringformat]l **** Idletag  *************************************/
            case 'l' : sprintf(cline+strlen(cline),u->idle ? fline : " "); break;
     
	    /***** %W[stringformat]L **** InfoLine *************************************/
	    case 'L' : sprintf(cline+strlen(cline),fline,u->user->info); break;

	    /***** %W[stringformat]I **** Ident ****************************************/
	    case 'I' : sprintf(cline+strlen(cline),fline,u->ident); break;

	    /***** %W[stringformat]i **** IP *******************************************/
	    case 'i' : get_hostname(u->ip,host); sprintf(cline+strlen(cline),fline,host); break;

	    /***** %W[stringformat]o **** Ownlogin *********************************/
            case 'o' : sprintf(cline+strlen(cline),u->pid == c->pid ? fline : " "); break;

	    /***** %W[stringformat]H **** Ident@IP *************************************/
	    case 'H' :
	      strncpy(host,u->ident,strlen(u->ident));
	      host[strlen(u->ident)]='@'; 
	      host[strlen(u->ident)+1]='\0'; 
	      get_hostname(u->ip,host+strlen(host)); 
	      sprintf(cline+strlen(cline),fline,host);
	      break;

	    default :
	      strncpy(cline+strlen(cline),line-strlen(fline)-3,strlen(fline)+3);
	      *(cline+strlen(cline)+strlen(fline)+4) = '\0';
	      break;
	  }
	  *fline='\0';
	  line++;
	}
      }
    } else
    { *(cline+strlen(cline)+1) = '\0';
      *(cline+strlen(cline)) = *line;
      line++;
    }
  }
  server_reply(c,cline);
  return 0;
}

int user_who(struct child_s* c, char** p, int n)
{ int             i;
  int             k=0;
  struct child_s* u;
  char*           p1;
  char*           p2;
  char            line[1024];

  prepare_msg(c,&who_top_msg);
  prepare_msg(c,&who_body_msg);
  prepare_msg(c,&who_tail_msg);
  /**********************************************/
  /* Cooking Who_Top! (Nothing to be cooked???) */
  /**********************************************/
  p1 = who_top_msg.msg;
  p2 = p1;
  do
  { while (*p2 && (*p2 != '\n')) p2++;
    if (*p2) p2++;
    line[0]='+';
    memcpy(line+1,p1,p2-p1);
    line[p2-p1+1]='\0';
    server_reply_who(c,line,c,0);
    p1 = p2;
  } while (*p2);
  /*********************/
  /* Cooking Who_Body! */
  /*********************/
  u = child;
  for (i=0; i<children; i++, u++) if (!u->admin && u->user &&
    !invisible(u->user))
  { p1 = who_body_msg.msg;
    p2 = p1;
    do
    { while (*p2 && (*p2 != '\n')) p2++;
      if (*p2) p2++;
      line[0]='+';
      memcpy(line+1,p1,p2-p1);
      line[p2-p1+1]='\0';
      server_reply_who(c,line,u,k+1);
      p1 = p2;
    } while (*p2);
    k++;
  }
  /***********************************************/
  /* Cooking Who_Tail! (Nothing to be cooked???) */
  /***********************************************/
  p1 = who_tail_msg.msg;
  p2 = p1;
  do
  { while (*p2 && (*p2 != '\n')) p2++;
    if (*p2) p2++;
    line[0]='+';
    memcpy(line+1,p1,p2-p1);
    line[p2-p1+1]='\0';
    server_reply_who(c,line,c,k);
    p1 = p2;
  } while (*p2);
  server_reply(c,"+");
  return 0;
}

int variable_expand(struct child_s* c, char** p, int n)
{ char str[128];

  switch (**p)
  { case 'a' :
      sprintf(str,"+%i",num_anon);
      break;
    case 'A' :
      if (limit) sprintf(str,"+%i",
        limit->max_sum ? limit->max_user : limit->max_anon);
      else strcpy(str,"+unlimited");
      break;
    case 'n' :
      sprintf(str,"+%i",num_anon+num_users);
      break;
    case 'N' :
      if (limit) sprintf(str,"+%i",
        limit->max_sum ? limit->max_user : limit->max_anon+limit->max_user);
      else strcpy(str,"+unlimited");
      break;
    case 'u' :
      sprintf(str,"+%i",num_users);
      break;
    case 'U' :
      if (limit) sprintf(str,"+%i",limit->max_user);
      else strcpy(str,"+unlimited");
      break;
    default :
      strcpy(str,"-?");
      break;
  }
  server_reply(c,str);
  return 0;
}

int f4adp_ident_check(struct child_s* c, char** p, int n)
{ struct sockaddr_in addr;
  struct hostent*    hp;
  unsigned long      ipaddr;

  if (!c->ident[0])
  { if (!sscanf(p[0],"%lu",&ipaddr)) ipaddr = 0;
    if (ipaddr) c->ip = ipaddr;
    if (reject_is_rejected_ip(c->ip,1))
    { server_reply(c,"-rejected ip");
      return 0;
    } else
    { strcpy(c->ident,p[1]);
      if (flag_get(flags,FLAG_PRIVATEMODE) && !in_idents(c->ip,c->ident))
      { server_reply(c,"-bad ident@ip");
        addr.sin_addr.s_addr = htonl(c->ip);
        hp = gethostbyaddr((char*)&addr.sin_addr,sizeof(addr.sin_addr),AF_INET);
        server_log_local(NULL,LG_LOGIN,"UNAUTHORIZED CONNECTION from %s (%s)",
          inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
        return 0;
      }
    }
  }
  server_reply(c,"+ok");
  return 0;
}

int f4adp_quit(struct child_s* c, char** p, int n)
{ child_logout(c,CL_NORMAL);
  return -1;
}

/* command table
   must be sorted alphabetically by command for binary search
*/
struct serverd_s
{ char  command[8];                              /* command string          */
  int   (*function)(struct child_s*,char**,int); /* command handler         */
  int   flags;                                   /* flags, see below        */
  int64 right;                                   /* right needed            */
  int   min_args;                                /* min number of arguments */
  int   max_args;                                /* max number of arguments */
  /* flags: 0x01 = must be logged in
            0x02 = perform permission check against right
            0x04 = for ftps only
  */
} serverd_cmd[] =
{ { "ADI", account_disable              , 0x03, R_USER_DISABLE   ,  1,  1 },
  { "AEN", account_enable               , 0x03, R_USER_DISABLE   ,  1,  1 },
  { "CFG_AVL", f4adp_cfg_avl            , 0x03, R_CONFIG         ,  0,  0 },
  { "CFG_GET", f4adp_cfg_get            , 0x03, R_CONFIG         ,  1,  1 },
  { "CFG_RM" , f4adp_cfg_rm             , 0x03, R_CONFIG         ,  1,  1 },
  { "CFG_SET", f4adp_cfg_set            , 0x03, R_CONFIG         ,  2,  2 },
  { "CNT", count                        , 0x01, 0                ,  0,  0 },
  { "COL", color                        , 0x01, 0                ,  0,  0 },
  { "DGF", dump_group_file              , 0x03, R_DUMP_FILES     ,  1,  1 },
  { "DMP_CH", f4adp_dump_children       , 0x01, 0                ,  0,  0 },
  { "DMP_HA", f4adp_dump_handler        , 0x01, 0                ,  0,  0 },
  { "DMP_LIM", f4adp_lim_dump           , 0x01, 0                ,  0,  0 },
  { "DMP_WA", f4adp_dump_watcher        , 0x01, 0                ,  0,  0 },
  { "DUF", dump_user_file               , 0x03, R_DUMP_FILES     ,  1,  1 },
  { "DUP", send_dupechk                 , 0x01, 0                ,  0,  0 },
  { "FCM", file_chmod                   , 0x05, 0                ,  1,  1 },
  { "FCO", file_chown                   , 0x05, 0                ,  1,  1 },
  { "FDN", file_downloaded              , 0x05, 0                ,  1,  1 },
  { "FRM", file_removed                 , 0x05, 0                ,  1,  1 },
  { "FRN", file_renamed                 , 0x05, 0                ,  1,  1 },
  { "FUP", file_uploaded                , 0x05, 0                ,  1,  1 },
  { "GRP", groups                       , 0x01, 0                ,  1,  2 },
  { "GRP_ADD", f4adp_grp_add            , 0x03, R_GROUP_ADD      ,  2,  2 },
  { "GRP_CHG", f4adp_grp_chg            , 0x03, R_GROUP_EDIT     ,  2, 99 },
  { "GRP_GET", f4adp_grp_get            , 0x03, R_GROUP_LIST     ,  1,  1 },
  { "GRP_LS" , f4adp_grp_ls             , 0x03, R_GROUP_LIST     ,  0,  0 },
  { "GRP_LSR", f4adp_grp_lsr            , 0x05, 0                ,  0,  0 },
  { "GRP_MDN", f4adp_grp_mdn            , 0x01, 0                ,  1,  1 },
  { "GRP_MUP", f4adp_grp_mup            , 0x01, 0                ,  1,  1 },
  { "GRP_RM" , f4adp_grp_rm             , 0x03, R_GROUP_DELETE   ,  1,  1 },
  { "GRP_SET", f4adp_grp_set            , 0x03, R_GROUP_EDIT     , 11, 11 },
  { "IDT_CHK", f4adp_ident_check        , 0x00, 0                ,  2,  2 },
  { "LIM_GBW", f4adp_lim_getbw          , 0x01, 0                ,  1,  1 },
  { "LIM_RBW", f4adp_lim_returnbw       , 0x01, 0                ,  0,  0 },
  { "LOG_ACT", f4adp_log_end            , 0x01, 0                ,  4,  4 },
  { "LOG_BEG", f4adp_log_begin          , 0x01, 0                ,  4,  4 },
  { "LOG_END", f4adp_log_end            , 0x01, 0                ,  4,  4 },
  { "MSG_AVL", f4adp_msg_avl            , 0x03, R_MESSAGES       ,  1,  1 },
  { "MSG_DFL", f4adp_msg_dfl            , 0x03, R_MESSAGES       ,  1,  1 },
  { "MSG_GET", f4adp_msg_get            , 0x03, R_MESSAGES       ,  2,  2 },
  { "MSG_RM" , f4adp_msg_rm             , 0x03, R_MESSAGES       ,  2,  2 },
  { "MSG_SET", f4adp_msg_set            , 0x03, R_MESSAGES       ,  3,  3 },
  { "PRT_GET", f4adp_port_get           , 0x03, R_PORTS          ,  0,  0 },
  { "PRT_SET", f4adp_port_set           , 0x03, R_PORTS          ,  2,  2 },
  { "QUIT",    f4adp_quit               , 0x00, 0                ,  0,  0 },
  { "REG_GET", f4adp_reg_get            , 0x03, R_REGISTRY       ,  1,  1 },
  { "REG_KEY", f4adp_reg_keys           , 0x03, R_REGISTRY       ,  0,  0 },
  { "REG_LS" , f4adp_reg_ls             , 0x03, R_REGISTRY       ,  0,  0 },
  { "REG_RM" , f4adp_reg_rm             , 0x03, R_REGISTRY       ,  1,  1 },
  { "REG_SET", f4adp_reg_set            , 0x03, R_REGISTRY       ,  2, 99 },
  { "SIN", send_startup_info            , 0x04, 0                ,  0,  0 },
  { "SRV_GST", f4adp_server_get_state   , 0x03, R_STATE          ,  0,  0 },
  { "SRV_SST", f4adp_server_set_state   , 0x03, R_STATE          ,  1,  1 },
  { "SSC", send_shortcut                , 0x01, 0                ,  0,  0 },
  { "STO", send_top                     , 0x01, 0                ,  0,  0 },
  { "UAN", user_anonymous               , 0x04, 0                ,  1,  1 },
  { "UIN", user_information             , 0x01, 0                ,  1,  1 },
  { "ULI", user_list                    , 0x01, 0                ,  0,  9 },
  { "ULL", user_list_loggedin           , 0x01, 0                ,  0,  1 },
  { "ULP", user_request_login_permission, 0x00, 0                ,  2,  3 },
  { "USR_ADD", f4adp_usr_add            , 0x03, R_USER_ADD       ,  1,  2 },
  { "USR_ADQ", f4adp_usr_adq            , 0x03, R_USER_ADD       ,  5, 14 },
  { "USR_CHG", f4adp_usr_chg            , 0x01, 0                ,  1, 99 },
  { "USR_GET", f4adp_usr_get            , 0x01, 0                ,  1,  1 },
  { "USR_KIC", f4adp_usr_kick           , 0x01, 0                ,  1,  2 },
  { "USR_KIL", f4adp_usr_kill           , 0x01, 0                ,  1,  2 },
  { "USR_LS" , f4adp_usr_ls             , 0x03, R_USER_READ      ,  0,  0 },
  { "USR_LSR", f4adp_usr_lsr            , 0x05, 0                ,  0,  0 },
  { "USR_PWD", f4adp_usr_pwd            , 0x01, 0                ,  1,  2 },
  { "USR_RM" , f4adp_usr_rm             , 0x01, 0                ,  1,  1 },
  { "USR_RST", f4adp_usr_rst            , 0x03, R_USER_EDIT_STATS,  1,  2 },
  { "USR_SET", f4adp_usr_set            , 0x01, 0                , 29, 29 },
  { "UUD", user_upload_download         , 0x04, 0                ,  1,  4 },
  { "UUR", user_upload_reduce           , 0x05, 0                ,  2,  2 },
  { "VAR_EXP", variable_expand          , 0x05, 0                ,  1,  1 },
  { "VER", server_version               , 0x01, 0                ,  0,  0 },
  { "VER_F4P", server_version_f4adp     , 0x01, 0                ,  0,  0 },
  { "WA_ADD", f4adp_watch_add           , 0x03, R_MONITOR        ,  2,  2 },
  { "WA_DEL", f4adp_watch_del           , 0x03, R_MONITOR        ,  0,  0 },
  { "WHO", user_who                     , 0x01, 0                ,  0,  0 } };

int child_communication(int fd, void* ptr)
{ struct child_s* c = ptr;
  char*           p[64];
  int             i;
  int             j;
  int             k;
  int             n;
  int             r;

  /* read command (scanner functionality) and handle errors */
  for (;;)
  { switch (read_command(c,p,60,&n))
    { case -2 :
        /* EOF */
        child_logout(c,CL_NORMAL);
        return 0;
      case -1 :
        /* error during fgets() */
        errlog("child_communication(): %s (%i), child_pid=%i, username=%s, logged out",
          strerror(errno),errno,c->pid,c->user ? c->user->name : "???");
        if (c->pid != -1) kill(c->pid,SIGKILL);
        child_logout(c,CL_PIPEERR);
        return 0;
      case  1 :
        errlog("child_communication(): too many parameters");
        server_reply(c,"-Too many parameters.");
        return 0;
      case  2 :
        errlog("child_communication(): command string of zero length received");
        server_reply(c,"-Command string of zero length received.");
        return 0;
      case  3 :
        errlog("child_communication(): buffer overrun");
        server_reply(c,"-Buffer overrun.");
        return 0;
      case  4 :
        errlog("child_communication(): no command in command string");
        server_reply(c,"-No command in command string.");
        return 0;
      case  5 :
        /* no or incomplete command, return and select() until more data arrives */
        return 0;
    }
    time(&c->act_time);
#ifdef DEBUG
    fprintf(debugf,"Child command (n=%i):",n);
    for (i=0; i<n; i++) fprintf(debugf," %s",*(p+i));
    fprintf(debugf,"\n");
#endif
    /* do binary search of command in command table */
    i = 0;
    k = (sizeof(serverd_cmd) / sizeof(struct serverd_s)) - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      r = strcmp(*p,serverd_cmd[j].command);
      if (r > 0)  i = j + 1; else if (r < 0) k = j - 1; else
      { /* command found, now check flags */
        r = serverd_cmd[j].flags;
        if (r & 0x01) if (!c->user)
        { server_reply(c,"-Permission denied (must be logged in)");
          break;
        }
        if (r & 0x02) if (!allowed(c->user,serverd_cmd[j].right,NULL))
        { server_reply(c,"-Permission denied");
          break;
        }
        if (r & 0x04) if (c->admin)
        { server_reply(c,"-Permission denied (ftps only)");
          break;
        }
        n--;
        if ((n < serverd_cmd[j].min_args) || (n > serverd_cmd[j].max_args))
        { server_reply(c,"-Syntax error (number of arguments)");
          break;
        }
        /* flag check ok, call command handler and return */
        j = serverd_cmd[j].function(c,p+1,n);
        if (j < 0) return -1;
        c -= j;
        break;
      }
    }
    if (i > k)
    { /* command not found in table */
      errlog("child_communication(): unknown client command (%s)",p[0]);
      server_reply(c,"-Unknown command");
    }
  }
  /* not reached */
  return 0;
}

/* EOF */
