/*   ftps.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include "globals.h"
#include "common.h"
#include "log.h"
#include "perm.h"
#include "rfc931.h"
#include "utils.h"

void quit(int reason)
{ close_server++;
  if (close_reason == CL_ERROR) close_reason = reason;
  if ((reason == CL_NORMAL) || (reason == CL_SOCKERR)) close_server++;
}

void signal_handler(int sig)
{ if (sig == SIGTERM) exit(CL_KILL);
  if (((sig_write==sig_size-1) && !sig_read) || (sig_write+1==sig_read))
    return;
  sig_array[sig_write] = sig;
  if (++sig_write >= sig_size) sig_write = 0;
}

void handle_signals(void)
{ while (sig_read != sig_write)
  { switch (sig_array[sig_read])
    { case SIGHUP :
        quit(CL_SHUTDOWN);
        break;
      case SIGINT :
        quit(CL_CLOSE);
        break;
      case SIGQUIT :
        quit(CL_KICK);
        break;
      case SIGPIPE :
        quit(CL_PIPEERR);
        break;
      case SIGALRM :
        quit(CL_TIMEOUT);
        break;
      case SIGUSR1 :
        close_reason = CL_ERROR;
        close_server = 0;
        break;
    }
    if (++sig_read >= sig_size) sig_read = 0;
  }
}

int cmd_abor();
int cmd_acct();
int cmd_allo();
int cmd_appe();
int cmd_cdup();
int cmd_cwd();
int cmd_dele();
int cmd_help();
int cmd_list();
int cmd_mdtm();
int cmd_mkd();
int cmd_mode();
int cmd_nlst();
int cmd_noop();
int cmd_pass();
int cmd_pasv();
int cmd_port();
int cmd_pwd();
int cmd_quit();
int cmd_rein();
int cmd_rest();
int cmd_retr();
int cmd_rmd();
int cmd_rnfr();
int cmd_rnto();
int cmd_site();
int cmd_size();
int cmd_smnt();
int cmd_stat();
int cmd_stor();
int cmd_stou();
int cmd_stru();
int cmd_syst();
int cmd_type();
int cmd_user();

int cmd_auth(char** p, int n, char* buf, int buflen)
{ quit(CL_NORMAL);
  return send_reply(421,"AUTH command out of sequence (bouncer timeout?)");
}

struct ftpcmd_s
{ char command[4];
  int  (*function)(char**,int,char*,int);
  int  must_be_loggedin;
  int  parse_command;
} ftpcmd[] =
{ { "ABOR", cmd_abor, 1, 0 }, { "ACCT", cmd_acct, 1, 0 },
  { "ALLO", cmd_allo, 1, 0 }, { "APPE", cmd_appe, 1, 0 },
  { "AUTH", cmd_auth, 0, 0 },
  { "CDUP", cmd_cdup, 1, 0 }, { "CWD",  cmd_cwd,  1, 0 },
  { "DELE", cmd_dele, 1, 0 }, { "HELP", cmd_help, 1, 0 },
  { "LIST", cmd_list, 1, 1 }, { "MDTM", cmd_mdtm, 1, 0 },
  { "MKD",  cmd_mkd,  1, 0 }, { "MODE", cmd_mode, 0, 0 },
  { "NLST", cmd_nlst, 1, 1 }, { "NOOP", cmd_noop, 1, 0 },
  { "PASS", cmd_pass, 1, 0 }, { "PASV", cmd_pasv, 1, 0 },
  { "PORT", cmd_port, 1, 0 }, { "PWD",  cmd_pwd,  1, 0 },
  { "QUIT", cmd_quit, 0, 0 }, { "REIN", cmd_rein, 1, 0 },
  { "REST", cmd_rest, 1, 0 }, { "RETR", cmd_retr, 1, 0 },
  { "RMD",  cmd_rmd,  1, 0 }, { "RNFR", cmd_rnfr, 1, 0 },
  { "RNTO", cmd_rnto, 1, 0 }, { "SITE", cmd_site, 1, 1 },
  { "SIZE", cmd_size, 1, 0 }, { "SMNT", cmd_smnt, 1, 0 },
  { "STAT", cmd_stat, 1, 0 }, { "STOR", cmd_stor, 1, 0 },
  { "STOU", cmd_stou, 1, 0 }, { "STRU", cmd_stru, 0, 0 },
  { "SYST", cmd_syst, 0, 0 }, { "TYPE", cmd_type, 0, 1 },
  { "USER", cmd_user, 0, 0 }, { "XMKD", cmd_mkd,  1, 0 },
  { "XPWD", cmd_pwd,  1, 0 }} ;

int dispatch_command(char* buffer, int bufferlen, char** p, int n)
{ int i = 0;
  int j;
  int k = (sizeof(ftpcmd) / sizeof(struct ftpcmd_s)) - 1;
  int r;

  while (i <= k)
  { j = (i + k) >> 1;
    r = strncasecmp(*p,ftpcmd[j].command,4);
    if (r < 0) k = j - 1; else if (r > 0) i = j + 1; else
    { if (!logged_in && ftpcmd[j].must_be_loggedin)
      { retries++;
        if (login_retries && (retries >= login_retries))
        { quit(CL_NORMAL);
          log_action(0,LG_LOGIN,"LOGIN","FAILED: Not logged in yet");
          return send_reply(421,"Not logged in yet, connection closed.");
        }
        return send_reply(503,"Please login first.");
      }
      if ((n == 2) && ftpcmd[j].parse_command)
      { n--;
        if (parse_command(*(p+1),p,31,&n))
          return send_reply(500,"Too many parameters.");
      }
      return ftpcmd[j].function(p+1,n-1,buffer,bufferlen);
    }
  }
  return send_message(500,1,"Unrecognised command (%s).",*p);
}

void get_string(char** buffer, char* str, int max)
{ fgets(str,max,stdin);
  if (*str)
  { if (!*buffer) *buffer = malloc(strlen(str)+1);
    if (!*buffer)
    { errlog("get_string(): out of memory");
      exit(close_reason);
    }
    if (str[strlen(str)-1] == '\n') str[strlen(str)-1] = '\0';
    strcpy(*buffer,str);
  } else *buffer = NULL;
}

void prepare_msg(struct msg_s* msg)
{ char*       buffer;
  int         err;
  FILE*       file;
  char        filename[256];
  char*       name;
  struct stat st;

  strcpy(filename,server_dir);
  buffer = filename + strlen(filename);
  if (usr.name[0])
  { strcpy(buffer,DFL_MSGUDIR);
    strcat(buffer,usr.name);
    strcat(buffer,"/");
    strcat(buffer,msg->filename);
    err = stat(filename,&st);
    if (err) name = get_groupname(usr.gid);
    if (err && name)
    { strcpy(buffer,DFL_MSGGDIR);
      strcat(buffer,name);
      strcat(buffer,"/");
      strcat(buffer,msg->filename);
      err = stat(filename,&st);
    }
  } else err = -1;
  if (err)
  { strcpy(buffer,DFL_MSGDIR);
    strcat(buffer,msg->filename);
    err = stat(filename,&st);
  }
  if (!err && (st.st_mtime != msg->filedate))
  { if (st.st_size+2 > msg->size)
    { buffer = alloc(msg->msg,st.st_size+2);
      if (buffer)
      { msg->msg = buffer;
        msg->size = st.st_size + 2;
      }
    } else buffer = msg->msg; 
    if (buffer)
    { file = fopen(filename,"rb");
      if (file)
      { msg->filedate = st.st_mtime;
        msg->len = fread(msg->msg,1,st.st_size,file);
        fclose(file);
        if (!msg->len || (*(msg->msg+msg->len-1) != '\n'))
        { *(msg->msg+msg->len) = '\n';
          msg->len++;
        }
        *(msg->msg+msg->len) = '\0';
      }
    }
  }
  if (!msg->msg || !msg->len)
  { msg->msg = msg->default_msg;
    msg->len = strlen(msg->msg);
    msg->size = 0;
  }
}

int main(int argc, char** argv, char** env)
{ int                bouncer;
  struct sockaddr_in loc_addr;
  int                n;
  char*              p[32];
#ifndef HAVE_SIGSET
  struct sigaction   sig_act;
#endif
  struct sockaddr_in rem_addr;
  char               str[256];
  int                timeout;

#ifndef HAVE_SIGSET
  sigfillset(&sig_act.sa_mask);
  sig_act.sa_flags = SA_RESTART;
#define sigset(x,y) sig_act.sa_handler = y; sigaction(x,&sig_act,NULL)
#endif
  sigset(SIGHUP,signal_handler);
  sigset(SIGINT,signal_handler);
  sigset(SIGQUIT,signal_handler);
  sigset(SIGALRM,signal_handler);
  sigset(SIGPIPE,signal_handler);
  sigset(SIGTERM,signal_handler);
  sigset(SIGUSR1,signal_handler);
  sigset(SIGUSR2,signal_handler);
  setvbuf(stdout,NULL,_IOLBF,0);
  printf("SIN\n");
  scanf("%li %i %i %li %i %i %i %i %i ",
    &other_ip_s,&login_retries,&flags,
    &ipaddr,&sock,&fconv_mode,&checktimeout,&bouncer,&timeout);
  get_string(&permission_file,str,sizeof(str));
  get_string(&readme_file,str,sizeof(str));
  get_string(&dupe_path,str,sizeof(str));
  get_string(&ff_path,str,sizeof(str));
  get_string(&pre_download,str,sizeof(str));
  get_string(&pre_upload,str,sizeof(str));
  fgets(server_dir,sizeof(server_dir),stdin);
  if (server_dir[strlen(server_dir)-1] == '\n')
    server_dir[strlen(server_dir)-1] = '\0';
  if ((sockf = fdopen(sock,"wt")) == NULL || (sockfr = fdopen(sock,"rt")) == NULL)
  { errlog("fdopen(): %s (%i)",strerror(errno),errno);
    return CL_SOCKERR;
  }
  setvbuf(sockf,NULL,_IOLBF,0);

  /* bouncer and ident check 0x01=bouncer, 0x02=ident */
  *usr.ident = '\0';
  usr.ipaddr = 0;
  if (bouncer & 0x01)
  { /* bouncer handling, bnc4all & rftp compatible
       bnc4all syntax: AUTH [<user>]@<ip>[ <anything_else_ignored>]
       rftp syntax:    IDNT [<user>]@<ip>
     */
    n = (int)fgets_to(str,sizeof(str),sockfr,timeout);
    if (n)
    { *p = strchr(str,'@');
      if (*p)
      { usr.ipaddr = ntohl(inet_addr((*p)+1));
        **p = '\0';
        *p = strchr(str,' ');
        if (*p)
        { strncpy(usr.ident,(*p)+1,sizeof(usr.ident));
          usr.ident[sizeof(usr.ident)-1] = '\0';
        }
      }
    }
  }
  if ((bouncer & 0x02) && !*usr.ident)
  { n = sizeof(loc_addr);
    if (getsockname(sock,(struct sockaddr*)&loc_addr,&n) < 0) bouncer = 0;
    n = sizeof(rem_addr);
    if (getpeername(sock,(struct sockaddr*)&rem_addr,&n) < 0) bouncer = 0;
    if (bouncer)
    { strncpy(usr.ident,rfc931(&rem_addr,&loc_addr),sizeof(usr.ident));
      usr.ident[sizeof(usr.ident)-1] = '\0';
    }
  }
  if (!*usr.ident) strcpy(usr.ident,"UNKNOWN");
  if (usr.ipaddr) ipaddr = usr.ipaddr;

  /* send ident@ip to ftpd for check, exit if check fails */
  printf("IDT_CHK %lu %s\n",usr.ipaddr,usr.ident);
  fgets(str,sizeof(str),stdin);
  if (*str != '+')
  { quit(CL_IDENT);
    printf("UUD %i\n",close_reason);
    fgets(str,sizeof(str),stdin);
    return close_reason;
  }

  /* send startup message and enter command loop */
  prepare_msg(&startup_msg);
  if (send_reply(220,startup_msg.msg) == -1) quit(CL_SOCKERR);
  while (close_server < 2)
  { handle_signals();
    if (close_server)
    { printf("UUD %i\n",close_reason);
      fgets(str,sizeof(str),stdin);
    }
    switch (read_ftp_command(str,sizeof(str),p,31,&n))
    { case -1 :
        if (errno)
        { if (errno == EINTR) handle_signals(); else
          { errlog("read_ftp_command(): %s (%i)",strerror(errno),errno);
            quit(CL_SOCKERR);
          }
        }
        break;
      case 1 :
        break;
      default :
        if (close_server)
        { close_server++;
          switch (close_reason)
          { case CL_CLOSE :
              send_reply(421,"Server closing, closing control connection");
              break;
            case CL_SHUTDOWN :
              send_reply(421,"Server shutdown, closing control connection");
              break;
            case CL_TIMEOUT :
              send_reply(421,"Timeout, closing control connection");
              break;
            case CL_KICK :
              send_reply(421,"You have been kicked, closing control connection");
              break;
          }
        } else
        { if (dispatch_command(str,sizeof(str),p,n) == -1)
            { if (errno == EINTR) handle_signals(); else quit(CL_SOCKERR); }
        }
    }
  }
  return close_reason;
}

/* EOF */
