/*   perm.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <utime.h>
#include <sys/stat.h>
#include "externs.h"
#include "perm.h"
#include "utils.h"

int superuser(void)
{ return (usr.class & UC_MASK) == UC_ROOT;
}

int groupop(void)
{ return (usr.class & UC_MASK) == UC_GROUPOP;
}

int siteop(void)
{ return (usr.class & UC_MASK) == UC_SITEOP;
}

int invisible(void)
{ return (usr.class & UF_INVISIBLE);
}

int coloruser(void)
{ return (usr.class & UF_COLOR);
}

int in_group(int32 gid)
{ int i;

  if (usr.gid == gid) return 1;
  for (i=0; i<8; i++) if ((usr.gids[i] != -1) && (usr.gids[i] == gid))
    return 1;
  return 0;
}

int32 get_gid(char* name)
{ struct group_s* g = group;
  int32           gid;
  int             i;

  for (i=0; i<group_count; i++)
  { if (!strcasecmp(g->name,name)) return g->gid;
    g++;
  }
  if (sscanf(name,SI32,&gid)) return gid; else return -1;
}

int32 get_uid(char* name)
{ int            i;
  struct user_s* u = user;
  int32          uid;

  for (i=0; i<user_count; i++)
  { if (!strcasecmp(u->name,name)) return u->uid;
    u++;
  }
  if (sscanf(name,SI32,&uid)) return uid; else return -1;
}

struct group_s* get_group_by_gid(int32 gid)
{ struct group_s* g = group;
  int             i;

  for (i=0; i<group_count; i++, g++) if (g->gid == gid) return g;
  return NULL;
}

char* get_groupname(int32 gid)
{ struct group_s* g;
  static char*    incoming = "incoming";
  static char*    nogroup = "nogroup";

  if (gid == -1) return nogroup;
  if (gid == -2) return incoming;
  g = get_group_by_gid(gid);
  if (g) return g->name; else return NULL;
}

struct user_s* get_user_by_uid(int32 uid)
{ int            i = 0;
  int            j;
  int            k = user_count - 1;

  while (i <= k)
  { j = (i + k) >> 1;
    if ((user+j)->uid < uid) i = j + 1;
      else if ((user+j)->uid > uid) k = j - 1; else
    { while (j && ((user+j-1)->uid == uid)) j--;
      return user+j;
    }
  }
  return NULL;
}

char* get_username(int32 uid)
{ static char*   incoming = "incoming";
  static char*   nobody = "nobody";
  struct user_s* u;

  if (uid == -1) return nobody;
  if (uid == -2) return incoming;
  u = get_user_by_uid(uid);
  if (u) return u->name; else return NULL;
}

void split_dir(char* fp, char** dir, char** name, int testdir)
{ char*       p;
  struct stat s;
  char        wd[256];

  if (testdir) if (!add_dir(cwd,fp,wd,sizeof(wd)) &&
    !stat(strcmp(wd,"/") ? wd+1 : ".",&s))
  { if (s.st_mode & S_IFDIR)
    { *dir = fp;
      *name = fp+strlen(fp);
      return;
    }
  }
  if (*fp == '/')
  { p = strrchr(fp+1,'/');
    if (!p)
    { *dir = NULL;
      *name = fp+1;
    } else
    { *p = '\0';
      *dir = fp;
      *name = p+1;
    }
  } else
  { p = strrchr(fp,'/');
    if (!p)
    { *dir = fp+strlen(fp);
      *name = fp;
    } else
    { *p = '\0';
      *dir = fp;
      *name =  p+1;
    }
  }
}

int add_dir(char* src, char* dir, char* dest, int maxlen)
{ char  c;
  char* p = dir;
  char* q;

  if (!dir)
  { strcpy(dest,"/");
    return 0;
  }
  if (*dir == '/')
  { strcpy(dest,"/");
    dir++;
    p++;
  } else
  { if (strlen(src) >= maxlen)
    { errno = ENAMETOOLONG;
      return -1;
    }
    if (dest != src) strcpy(dest,src);
  }
  while (*dir)
  { while (*p && *p!='/') p++;
    c = *p;
    *p = '\0';
    if (!strcmp(dir,".."))
    { if (!*dest || !strcmp(dest,"/"))
      { errno = ENOENT;
        return -1;
      }
      q = dest+strlen(dest)-1;
      while (q>dest && *q!='/') q--;
      if ((q == dest) && (*q == '/')) *(q+1) = '\0'; else *q = '\0';
    } else if (strcmp(dir,".") && strcmp(dir,"/"))
    { if (strlen(dest)+strlen(dir)+1 >= maxlen)
      { errno = ENAMETOOLONG;
        return -1;
      }
      if (strcmp(dest,"/")) strcat(dest,"/");
      strcat(dest,dir);
    }
    if (c) *(p++) = c;
    dir = p;
  }
  return 0;
}

FILE* file_open(char* dir, char* filename, const char* mode, int32 pcheck)
{ struct stat    st;
  struct utimbuf t;
  char           wd[256];

  if (add_dir(cwd,dir,wd,sizeof(wd)) == -1) return NULL;
  if (strlen(wd)+strlen(filename)+1 >= sizeof(wd))
  { errno = ENAMETOOLONG;
    return NULL;
  }
  if (strcmp(wd,"/")) strcat(wd,"/");
  strcat(wd,filename);
  if (pcheck && !superuser())
  { if (!strcmp(filename,permission_file))
    { errno = EACCES;
      return NULL;
    }
    if (readme_file && !strcmp(filename,readme_file))
    { errno = EACCES;
      return NULL;
    }
    if (pcheck == P_READ)
    { if (stat(wd+1,&st) == -1) return NULL;
      if (permission_ok(dir,P_CHDIR) == -1) return NULL;
      if (file_permission_ok(dir,filename,FP_READ) == -1) return NULL;
    } else
    { if (stat(wd+1,&st) == -1)
      { if (permission_ok(dir,P_WRITE) == -1) return NULL;
      } else
      { if (permission_ok(dir,P_CHDIR) == -1) return NULL;
        if (file_permission_ok(dir,filename,FP_WRITE) == -1) return NULL;
      }
    }
  }
  if (flag_get(flags,FLAG_TOUCH))
  { t.modtime = time(&t.actime);
    utime(wd+1,&t);
  }
  return fopen(wd+1,mode);
}

char* find_cdpath(char* dir)
{ int  l;
  int  ofs;
  char wd[256];

  if (!add_dir(cwd,dir,wd,sizeof(wd)))
  { ofs = 0;
    while (ofs < cdpath_len)
    { l = strlen(cdpath+ofs);
      if (!strncmp(wd,cdpath+ofs,l) && ((wd[l] == '\0') || (wd[l] == '/')))
      { ofs += (l + 8) & ~0x07;
        return cdpath+ofs;
      }
      ofs += ((l + 8) & ~0x07) + (sizeof(int32) * 3);
    }
  }
  return NULL;
}

int read_cdpath(char* dir, int32* uid, int32* gid, int32* perm)
{ char* cdp;

  cdp = find_cdpath(dir);
  if (cdp)
  { *uid = *(int32*)cdp;
    *gid = *(int32*)(cdp+sizeof(int32));
    *perm = *(int32*)(cdp+(sizeof(int32)*2));
    return 0;
  }
  return -1;
}

int read_permission_file(char* dir, int32* uid, int32* gid, int32* perm,
  int32* dlc, char** buffer, int* n)
{ FILE*         file;
  char          line[32];
  struct perm_s p;
  struct stat   st;

  if (dlc) *dlc = 1;
  if (buffer) *buffer = NULL;
  if (n) *n = 0;
  if (read_cdpath(dir,uid,gid,perm) == 0) return 0;
  if ((file = file_open(dir,permission_file,"rb",0)) == NULL) return -1;
  if (fstat(fileno(file),&st) == -1)
  { fclose(file);
    return -1;
  }
  if (st.st_size < sizeof(struct perm_s))
  { if (buffer && n)
    { *n = 0;
      if ((*buffer = malloc(sizeof(struct file_s))) == NULL)
      { fclose(file);
        return -1;
      }
    }
    if (fread(line,1,st.st_size,file) != st.st_size)
    { if (buffer) free(*buffer);
      fclose(file);
      return -1;
    }
    if (sscanf(line,SI32" "SI32" "SI32,uid,gid,perm) != 3)
    { if (buffer) free(*buffer);
      fclose(file);
      return -1;
    }
    if (dlc) *dlc = 0;
  } else
  { if (!fread(&p,sizeof(p),1,file))
    { fclose(file);
      return -1;
    }
    *uid = p.uid;
    *gid = p.gid;
    *perm = p.perm;
    if (dlc) *dlc = p.dlc;
    if (buffer && n)
    { st.st_size -= sizeof(p);
      if ((*buffer = malloc(st.st_size+sizeof(struct file_s))) == NULL)
      { fclose(file);
        return -1;
      }
      if (fread(*buffer,1,st.st_size,file) != st.st_size)
      { free(*buffer);
        fclose(file);
        return -1;
      }
      *n = st.st_size / sizeof(struct file_s);
    }
  }
  fclose(file);
  return 0;
}

int remove_permission_file(char* dir)
{ char wd[256];

  if (strlen(dir)+strlen(permission_file) >= sizeof(wd))
  { errno = ENAMETOOLONG;
    return -1;
  }
  if (strcmp(dir,"/")) sprintf(wd,"%s/%s",dir+1,permission_file); else
    strcpy(wd,permission_file);
  unlink(wd);
  if (readme_file)
  { if (strcmp(dir,"/")) sprintf(wd,"%s/%s",dir+1,readme_file); else
      strcpy(wd,readme_file);
    unlink(wd);
  }
  return 0;
}

int write_permission_file(char* dir, int32 uid, int32 gid, int32 perm)
{ int           e;
  FILE*         file;
  struct perm_s p;
  struct stat   st;

  if (find_cdpath(dir)) return 0;
  if ((file = file_open(dir,permission_file,"r+b",0)) == NULL)
    if ((file = file_open(dir,permission_file,"w+b",0)) == NULL) return -1;
  if (fstat(fileno(file),&st) == -1) st.st_size = 0;
  p.uid = uid;
  p.gid = gid;
  p.perm = perm;
  if (st.st_size < sizeof(p))
  { p.dlc = p.dlct = 0;
    memset(&p.reserved,0,sizeof(p.reserved));
  }
  if (!fwrite(&p,st.st_size < sizeof(p) ? sizeof(p) : sizeof(int32)*3,1,file))
  { e = errno;
    fclose(file);
    errno = e;
    return -1;
  }
  fclose(file);
  return 0;
}

/* returns:
       -1 ... permission denied
     >= 0 ... ok (returns directory flags)
 */
int permission_part_ok(char* dir, int32 p, int uidok)
{ int32 gid;
  int32 perm;
  int32 uid;

  if (read_permission_file(dir,&uid,&gid,&perm,NULL,NULL,NULL) == -1)
  { if (superuser()) return 0;
    errno = EACCES;
    return -1;
  }
  if ((p != P_NOLIMIT) && superuser()) return (perm >> 24) & 0xFF;
/*  if (uidok) if (usr.uid == uid) return (perm >> 24) & 0xFF;*/
  errno = EACCES;
  if (p & P_CHMOD) if (usr.uid != uid) return -1;
  if (usr.uid == uid) if (((perm >> 16) & p) == p) return (perm >> 24) & 0xFF;
  if (in_group(gid)) if (((perm >> 8) & p) == p) return (perm >> 24) & 0xFF;
  if ((perm & p) == p) return (perm >> 24) & 0xFF;
  return -1;
}

int directory_hidden(char* dir)
{ int32 gid;
  int32 perm;
  int32 uid;

  if (superuser()) return 0;
  if (read_permission_file(dir,&uid,&gid,&perm,NULL,NULL,NULL) < 0) return 0;
  if ((perm >> 24) & DP_HIDDEN)
  { if (usr.uid == uid) if ((perm >> 16) & P_CHDIR) return 0;
    if (in_group(gid)) if ((perm >> 8) & P_CHDIR) return 0;
    if (perm & P_CHDIR) return 0;
    return 1;
  }
  return 0;
}

int directory_flag_set(char* dir, int flag)
{ int32 gid;
  int32 perm;
  int32 uid;

  if (read_permission_file(dir,&uid,&gid,&perm,NULL,NULL,NULL) < 0) return 0;
  if ((perm >> 24) & flag) return 1;
  return 0;
}

int permission_ok(char* dir, int32 perm)
{ char        c;
  int         n;
  char*       p;
  struct stat st;
  char        wd[256];

  if ((perm<0) || (perm>255))
  { errno = EACCES;
    return -1;
  }
  if (!dir) strcpy(wd,"/"); else
    if (add_dir(cwd,dir,wd,sizeof(wd)) == -1) return -1;
  if (strcmp(wd,"/"))
  { if (stat(wd+1,&st) == -1) return -1;
    if (!S_ISDIR(st.st_mode))
    { errno = ENOTDIR;
      return -1;
    }
  }
  if (superuser()) return permission_part_ok(wd,perm,0);
  p = wd+1;
  do
  { while (*p && (*p != '/')) p++;
    c = *p;
    *p = '\0';
    if (c && *(p+1))
    { n = permission_part_ok(wd,P_CHDIR,0);
      *(p++) = c;
      if (n == -1) return -1;
    } else
    { n = permission_part_ok(wd,perm,0);
      *p = c;
      if (n == -1) return -1;
      if (c) p++;
    }
  } while (*p);
  return n;
}

struct file_s* find_name(struct file_s* buffer, char* name, int n, int* found)
{ int i = 0;
  int j;
  int k = n-1;

  while (i <= k)
  { j = (i + k) >> 1;
    n = strcmp(name,(buffer+j)->name);
    if (n < 0) k = j - 1; else if (n > 0) i = j + 1; else
    { *found = 1;
      return buffer+j;
    }
  }
  *found = 0;
  return buffer+i;
}

int32 dirperm_to_fileperm(int32 perm) 
{ int   i;
  int32 p = 0x01000000;

  for (i=0; i<=2; i++)
  { if ((perm >> ((2-i)*8)) & P_READ) p |= FP_READ << (2-i)*8;
    if ((perm >> ((2-i)*8)) & P_OVERWRITE) p |= FP_WRITE << (2-i)*8;
    if (((perm >> ((2-i)*8)) & P_CHMOD) && !i) p |= FP_CHMOD << (2-i)*8;
  }
  return p;   
}

int read_file_permission(char* dir, char* name, int32* uid, int32* gid,
  int32* perm)
{ char*          buffer;
  struct file_s* f;
  int            found;
  int            n;

  if (read_permission_file(dir,uid,gid,perm,NULL,&buffer,&n) == -1)
  { errno = EACCES;
    return -1;
  }
  f = find_name((struct file_s*)buffer,name,n,&found);
  if (found)
  { *uid = f->uid;
    *gid = f->gid;
    if (*perm) *perm = f->perm; else *perm = dirperm_to_fileperm(*perm);
  } else *perm = dirperm_to_fileperm(*perm);
  free(buffer);
  return 0;
}

int file_permission_ok(char* dir, char* name, int32 tperm)
{ int32 gid;
  int32 perm;
  int32 uid;

  if (*name && (!strcmp(permission_file,name) || !strcmp(readme_file,name)))
  { errno = EACCES;
    return -1;
  }
  if (superuser()) return 0;
  if (read_file_permission(dir,name,&uid,&gid,&perm) == -1) return -1;
  if (uid == usr.uid) if ((((perm >> 16) & FP_ALL) & tperm) == tperm)
    return 0;
  if (in_group(gid)) if ((((perm >> 8) & FP_ALL) & tperm) == tperm)
    return 0;
  if (((perm & FP_ALL) & tperm) == tperm) return 0;
  errno = EACCES;
  return -1;
}

/* EOF */
