/*   transfer.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-2000 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include "tweak.h"
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include "externs.h"
#include "common.h"
#include "ff.h"
#include "list.h"
#include "log.h"
#include "perm.h"
#include "transfer.h"
#include "utils.h"

void alarmhandler(int sig)
{ scriptabort = 1;
  longjmp(scriptabortjmp,1);
}

int cmd_mode(char** p, int n)
{ if (n != 1) return send_reply(501,"Syntax: MODE <mode_id>");
  if (!strcasecmp(*p,"S")) return send_reply(200,"Mode set to STREAM.");
  if (!strcasecmp(*p,"B")) return send_reply(504,"Block mode not supported.");
  if (!strcasecmp(*p,"C")) return send_reply(504,"Compressed mode not supported.");
  return send_reply(501,"Invalid mode identifier.");
}

int cmd_stru(char** p, int n)
{ if (n != 1) return send_reply(501,"Syntax: STRU <structure_id>");
  if (!strcasecmp(*p,"F"))
  { structure = S_FILE;
    return send_reply(200,"Structure set to FILE.");
  }
  if (!strcasecmp(*p,"R"))
  { structure = S_RECORD;
    return send_reply(200,"Structure set to RECORD.");
  }
  if (!strcasecmp(*p,"P"))
    return send_reply(504,"Page structure not supported.");
  return send_reply(501,"Invalid structure identifier.");
}

int cmd_type(char** p, int n)
{ if (n<1 || n>2) return send_reply(501,"Syntax: TYPE <type_id>");
  if (!strcasecmp(*p,"A"))
  { type = T_ASCII;
    return send_reply(200,"Type set to ASCII.");
  }
  if (!strcasecmp(*p,"I"))
  { type = T_IMAGE;
    return send_reply(200,"Type set to IMAGE.");
  }
  if (!strcasecmp(*p,"E")) return send_reply(504,"EBCDI type not supported.");
  if (!strcasecmp(*p,"L")) return send_reply(504,"LOCAL type not supported.");
  return send_reply(501,"Invalid type identifier.");
}

int cmd_mdtm(char** p, int n)
{ char        dir[256];
  struct stat st;

  if (n != 1) return send_reply(501,"Syntax: MDTM <filename>");
  if ((add_dir(cwd,*p,dir,sizeof(dir)) == -1) || (stat(dir+1,&st) == -1))
    return send_error(550,"mdtm",errno);
  if (st.st_mode & S_IFREG)
  { strftime(dir,sizeof(dir),"%Y%m%d%H%M%S",localtime(&st.st_mtime));
    return send_reply(213,dir);
  } else return send_message(550,1,"%s: not a plain file.",*p);
}

int cmd_size(char** p, int n)
{ char        dir[256];
  struct stat st;

  if (n != 1) return send_reply(501,"Syntax: SIZE <filename>");
  if ((add_dir(cwd,*p,dir,sizeof(dir)) == -1) || (stat(dir+1,&st) == -1))
    return send_error(550,"size",errno);
  if (st.st_mode & S_IFREG)
  { sprintf(dir,"%u",(unsigned int)st.st_size);
    return send_reply(213,dir);
  } else return send_message(550,1,"%s: not a plain file.",*p);
}

int cmd_rest(char** p, int n)
{ int se = 0;

  if (type == T_ASCII)
    return send_reply(502,"REST not possible with ASCII type.");
  if (n == 1)
  { if (!sscanf(*p,"%li",&rest_offset)) se = -1;
  } else se = -1;
  if (se == -1) return send_reply(501,"Syntax: REST <offset>");
  return send_reply(350,"REST command successful.");
}

int limit_ok(unsigned long size)
{ if (superuser() || !usr.limit) return 1;
  return size <= usr.credit;
}

int cmd_transferpermitted(char* prog,char* buffer)
{ char  str[256];
  int   ec;
  char* p[8];

  p[0] = prog;
  p[1] = buffer;

  if (run_prog_xt(prog,p,&ec,str) == -1) return 0;

  if (WIFEXITED(ec))
  { if (WEXITSTATUS(ec) && (WEXITSTATUS(ec) != -1)) ec = 550; else ec = 150;
  } else ec = 150;
  if (ec == 550)
  { sprintf(buffer,"Transfer initial check failed (%s).",str);
    if (*buffer) if (send_message(ec,ec == 550,buffer,NULL) == -1) return -1;
  }
  return ec == 550 ? 1 : 0;
}

int cmd_retrieve(char* buffer, int buffersize, char* dir, char* name,
  char* dirandname, FILE* file, FILE** sockptr, int max_speed, char** p)
{ unsigned long      down = 0;
  int                err;
  fd_set             fdset_r;
  fd_set             fdset_w;
  int                fd_ctrl;
  int                fd_file;
  int                fd_sock;
  int                hidden;
  int                len;
  int                limited;
  int                maxfd;
  int                min_speed;
  int                n;
  FILE*              sock = *sockptr;
  struct sockaddr_in srv_addr;
  struct stat        st;
  struct timeval     t;
  time_t             t1;
  time_t             t2;
  time_t             t3;

  min_speed = dl_speed & 0xFFFF;
  if ((max_speed > 0) && (min_speed >= max_speed)) min_speed = max_speed - 1;
  if ((type == T_IMAGE) && rest_offset)
  { if (fseek(file,rest_offset,SEEK_SET) < 0)
      return send_error(550,"seek",errno);
    rest_offset = 0;
  }
  limited = permission_part_ok(dir,P_NOLIMIT,1) < 0;
  if (limited && (usr.limit < 0))
    return send_reply(550,"You have no download permission");
  if (fstat(fileno(file),&st) == -1) st.st_size = 0;
  if (!(st.st_mode & S_IFREG))
    return send_message(550,1,"%s: not a plain file.",name);
  if (limited && !limit_ok(st.st_size))
  { sprintf(buffer,"Not enough credit (you have: "SI64" KB / you need: %u KB)",
      usr.credit/1024,(unsigned int)st.st_size/1024);
    return send_reply(550,buffer);
  }
  if (pasvsock == -1)
  { if ((fd_sock = socket(PF_INET,SOCK_STREAM,0)) < 0)
      return send_error(425,"socket",errno);
    if (other_ip_s)
    { memset(&srv_addr,0,sizeof(srv_addr));
      srv_addr.sin_family = AF_INET;
      srv_addr.sin_port = 0;
      srv_addr.sin_addr.s_addr = other_ip_s;
      if (bind(n,(struct sockaddr*)&srv_addr,sizeof(srv_addr)) < 0)
        return send_error(425,"bind",errno);
    }
    if (connect_to(fd_sock,(struct sockaddr*)&addr,sizeof(addr),30) < 0)
      return send_error(425,"connect",errno);
  } else
  { len = sizeof(addr);
    fd_sock = accept_to(pasvsock,(struct sockaddr*)&addr,&len,30);
    if (fd_sock < 0) return send_error(425,"accept",errno);
  }
  if (type == T_ASCII)
  { err = fcntl(fd_sock,F_GETFL);
    if (err >= 0) err = fcntl(fd_sock,F_SETFL,err | O_NONBLOCK);
    if (err < 0)
    { err = errno;
      close(fd_sock);
      return send_error(425,"fcntl",err);
    }
  }
  sock = fdopen(fd_sock,type == T_ASCII ? "wt" : "wb");
  if (!sock)
  { err = errno;
    close(fd_sock);
    return send_error(425,"fdopen",err);
  }
  sprintf(buffer,"Opening %s data connection for %s (%lu bytes).",
    type == T_ASCII ? "ASCII" : "BINARY", name,(unsigned long)st.st_size);
  if (send_reply(150,buffer) == -1) return -1;
  hidden = directory_flag_set(dir,DP_HIDDEN);
  *(name-1) = '/';
  add_dir(cwd,dirandname,buffer,buffersize);
  log_begin(hidden,LG_GET,"DOWNLOAD","%s (%lu bytes)",buffer,
    (unsigned long)st.st_size);
  *(name-1) = '\0';
  time(&t1);
  FD_ZERO(&fdset_r);
  FD_ZERO(&fdset_w);
  fd_ctrl = fileno(sockfr);
  fd_file = fileno(file);
  maxfd = fd_ctrl > fd_sock ? fd_ctrl + 1 : fd_sock + 1;
  n = 1;
  err = 0;
  while (n > 0)
  { FD_SET(fd_ctrl,&fdset_r);
    FD_SET(fd_sock,&fdset_w);
    t.tv_sec  = timeout;
    t.tv_usec = 0;
    n = select(maxfd,&fdset_r,&fdset_w,NULL,&t);
    if (!n)
    { send_reply(426,"Timeout, data connection closed");
      n = -1;
      break;
    }
    if (n < 0)
    { send_error(426,"select",err=errno);
      break;
    }
    if (FD_ISSET(fd_ctrl,&fdset_r))
    { switch (read_ftp_command(buffer,buffersize,p,4,&n))
      { case 0 :
          if (strcasecmp(*p,"ABOR"))
          { if (send_reply(503,"Unexpected command.") < 0) break;
          } else
          { send_reply(426,"Transfer interrupted.");
            send_reply(226,"Transfer successfully aborted.");
          }
      }
      n = -1;
    } else
    { if (type == T_ASCII)
      { while (fgets(buffer,buffersize-2,file))
        { n = strlen(buffer);
          if (((n == 1) && (*buffer == '\n')) || ((n >= 2) &&
            (buffer[n-1] == '\n') && (buffer[n-2] != '\r')))
          { buffer[n-1] = '\r';
            buffer[n] = '\n';
            buffer[n+1] = '\0';
          }
          if (n) if (fputs(buffer,sock) < 0)
          { if (errno == EAGAIN) fseek(file,-n,SEEK_CUR); else
            { err = errno;
              n = -1;
            }
            break;
          }
          down += n;
        }
        if (n >= 0)
        { if (feof(file)) n = 0; else
          { if (errno == EAGAIN)
            { down--;
              n = 1;
            } else n = -1;
          }
        }
      } else
      { n = read(fd_file,buffer,buffersize);
        if (n > 0) n = write(fd_sock,buffer,n);
      }
      if (n < 0) send_error(426,"send",err=errno); else
      if (n > 0)
      { down += n;
        /* enforce minimum and maximum speed, t2=time elapsed, t3=time limit */
        if (min_speed || max_speed)
        { t2 = time(NULL) - t1;
          if (min_speed && (t2 >= 30))
          { t3 = (down/1024) / min_speed;
            if (t2 > t3)
            { sprintf(buffer,"Minimum speed of %i kb/s not reached, data connection closed",min_speed);
                send_reply(426,buffer);
                n = -1;
            }
          }
          if (max_speed)
          { t3 = (down/1024) / max_speed;
            if (t2 < t3)
            { t3 -= t2;
              sleep(t3 > 5 ? 5 : t3);
            }
          }
        }
      }
    }
  }
  fclose(sock);
  sock = NULL;
  time(&t2);
  t2 -= t1;
  t1 = t2;
  if (!t2) t2 = 1;
  t2 = down / t2;
  last_speed = t2;
  if (limited)
  { bytes_downloaded += down;
    printf("UUD 0 %lu 0 %li\n",down,t1);
    scanf(SI64,&usr.credit);
    fgetc(stdin);
  }
  len = find_cdpath(dir) == NULL;
  *(name-1) = '/';
  add_dir(cwd,dirandname,buffer,buffersize);
  if (err) log_end(hidden,LG_GET,"DOWNLOAD","%s FAILED: %s",buffer,strerror(err));
    else log_end(hidden,LG_GET,"DOWNLOAD","%s (%lu bytes at %3.1f Kb/s)",
      buffer,down,(float)t2/1024.0);
  if (len)
  { printf("#1#FDN %s\n",buffer);
    fgets(buffer,buffersize,stdin);
  }
  return err ? 0 : send_message(226,1,download_msg.msg,name);
}

int cmd_retr(char** p, int n)
{ char  buffer[8192];
  char* dir;
  FILE* file;
  FILE* sock = NULL;
  char* name;

  if (n != 1) return send_reply(501,"Syntax: RETR <filename>");
  n = add_dir(cwd,*p,buffer,sizeof(buffer));
  split_dir(*p,&dir,&name,0);
  if (*pre_download) switch (cmd_transferpermitted(pre_download,buffer))
  { case -1 : return -1;
    case  1 : return 0;
  }
  file = file_open(dir,name,type == T_ASCII ? "rt" : "rb",P_READ);
  if (!file) return send_error(550,"read",errno);
  printf("LIM_GBW 0\n");
  if (!fgets(buffer,sizeof(buffer),stdin))
  { n = errno;
    fclose(file);
    return send_error(550,"pipe error",n);
  }
  if (*buffer == '+')
  { if (!sscanf(buffer+1,"%i",&n)) n = 0;
  } else n = 0;
  if (!n)
  { fclose(file);
    return send_reply(550,"transfer not permitted (no bandwidth)");
  }
  if (n < 0) n = 0;
  n = cmd_retrieve(buffer,sizeof(buffer),dir,name,*p,file,&sock,n,p);
  fclose(file);
  if (sock) fclose(sock);
  printf("LIM_RBW\n");
  fgets(buffer,sizeof(buffer),stdin);
  return n;
}

int cmd_stor_check(char* buffer, char* dir, char* name)
{ struct check_s* ch;
  int             i;
  char*           p[8];
  void            (*oldvec)() = SIG_IGN;

  for (i=0, ch=check; i<check_count; i++, ch++)
    if (wildcard_matching(ch->wildcard,name,0))
  { if (add_dir(cwd,dir,buffer,512) == -1) return 0;
    i = 0;
    p[i++] = ch->name;
    if (*ch->param) p[i++] = ch->param;
    p[i++] = name;
    p[i++] = usr.name;
    if (checktimeout)
    { if (setjmp(scriptabortjmp))
      { if (scriptabort == 1)
        { scriptabort = 0;
          return 2;
        }
      }
      oldvec = signal(SIGALRM,alarmhandler);
      alarm(checktimeout);
    }
    if (run_program(ch->name,p,i,0,0,buffer+1,&i,buffer) == -1) return 0;
    if (checktimeout)
    { alarm(0);
      signal(SIGALRM,oldvec);
    }
    if (WIFEXITED(i))
    { if (WEXITSTATUS(i) && (WEXITSTATUS(i) != -1)) i = 550; else i = 226;
    } else i = 226;
    if ((i == 550) && !*buffer) strcpy(buffer,"File is corrupt.");
    if (*buffer) if (send_message(i,i == 550,buffer,NULL) == -1) return -1;
    if (i == 550)
    { add_dir(cwd,dir,buffer,512);
      add_dir(buffer,name,buffer,512);
      unlink(buffer+1);
    }
    return i == 550 ? 1 : 0;
  }
  return 0;
}

int cmd_stor_dcheck(char* buffer, char* name)
{ struct check_s* ch = dcheck;
  int             i;
  char*           p[8];

  for (i=0; i<dcheck_count; i++, ch++)
    if (wildcard_matching(ch->wildcard,name,0))
  { i = 0;
    p[i++] = ch->name;
    if (*ch->param) p[i++] = ch->param;
    p[i++] = buffer + 1;
    p[i++] = usr.name;
    if (run_program(ch->name,p,i,0,0,NULL,&i,buffer) == -1) return 0;
    if (WIFEXITED(i))
    { if (WEXITSTATUS(i) && (WEXITSTATUS(i) != -1)) i = 550; else i = 150;
    } else i = 150;
    if ((i == 550) && !*buffer) strcpy(buffer,"Dupe check failed.");
    if (*buffer) if (send_message(i,i == 550,buffer,NULL) == -1) return -1;
    return i == 550 ? 1 : 0;
  }
  return 0;
}

void cmd_stor_rm(char* fullname, char* dir, char* name)
{ if (add_dir(cwd,dir,fullname,512) == -1) return;
  if (add_dir(fullname,name,fullname,512) == -1) return;
  unlink(fullname+1);
}

int cmd_store(char* buffer, int buffersize, char* dir, char* name,
  char* dirandname, FILE* file, FILE** sockptr, int max_speed,
  char** p, struct stat* st)
{ int            err;
  fd_set         fdset_r;
  int            fd_ctrl;
  int            fd_file;
  int            fd_sock;
  int            hidden;
  int            limited;
  int            maxfd;
  int            min_speed;
  int            n;
  int            ok;
  FILE*          sock = *sockptr;
  struct timeval t;
  time_t         t1;
  time_t         t2;
  time_t         t3;
  unsigned int   up = 0;

  min_speed = ul_speed & 0xFFFF;
  if ((max_speed > 0) && (min_speed >= max_speed)) min_speed = max_speed - 1;
  limited = permission_part_ok(dir,P_NOLIMIT,0) < 0;
  if (pasvsock == -1)
  { if ((fd_sock = socket(PF_INET,SOCK_STREAM,0)) < 0)
      return send_error(425,"socket",errno);
    if (connect_to(fd_sock,(struct sockaddr*)&addr,sizeof(addr),30) < 0)
      return send_error(425,"connect",errno);
  } else
  { ok = sizeof(addr);
    fd_sock = accept_to(pasvsock,(struct sockaddr*)&addr,&ok,30);
    if (fd_sock < 0) return send_error(425,"accept",errno);
  }
  if (type == T_ASCII)
  { err = fcntl(fd_sock,F_GETFL);
    if (err >= 0) err = fcntl(fd_sock,F_SETFL,err | O_NONBLOCK);
    if (err < 0)
    { err = errno;
      close(fd_sock);
      return send_error(425,"fcntl",err);
    }
  }
  sock = fdopen(fd_sock,type == T_ASCII ? "rt" : "rb");
  if (!sock)
  { ok = errno;
    close(fd_sock);
    return send_error(425,"fdopen",ok);
  }
  sprintf(buffer,"Opening %s data connection for %s",
    type == T_ASCII ? "ASCII" : "BINARY",name);
  if (send_reply(150,buffer) == -1) return -1;
  hidden = directory_flag_set(dir,DP_HIDDEN);
  *(name-1) = '/';
  add_dir(cwd,dirandname,buffer,buffersize);
  log_begin(hidden,LG_PUT,"UPLOAD","%s",buffer);
  *(name-1) = '\0';
  time(&t1);
  FD_ZERO(&fdset_r);
  fd_ctrl = fileno(sockf);
  fd_file = fileno(file);
  maxfd = fd_ctrl > fd_sock ? fd_ctrl + 1 : fd_sock + 1;
  n = 1;
  err = 0;
  while (n > 0)
  { FD_SET(fd_ctrl,&fdset_r);
    FD_SET(fd_sock,&fdset_r);
    t.tv_sec  = timeout;
    t.tv_usec = 0;
    n = select(maxfd,&fdset_r,NULL,NULL,&t);
    if (!n)
    { send_reply(426,"Timeout, data connection closed");
      break;
    }
    if (n < 0)
    { send_error(426,"select",err=errno);
      n = -1;
      break;
    }
    if (FD_ISSET(fd_ctrl,&fdset_r))
    { switch (read_ftp_command(buffer,buffersize,p,4,&n))
      { case  0 :
          if (strcasecmp(*p,"ABOR"))
          { send_reply(503,"Unexpected command.");
          } else
          { send_reply(426,"Transfer interrupted.");
            send_reply(226,"Transfer successfully aborted.");
          }
      }
      n = -1;
    } else
    { if (type == T_ASCII)
      { while (fgets(buffer,buffersize,sock))
        { n = strlen(buffer);
          if ((n >= 2) && (buffer[n-1] == '\n') && (buffer[n-2] == '\r'))
          { buffer[--n] = '\0';
            buffer[n-1] = '\n';
          }
          if (n) if (fputs(buffer,file) < 0)
          { send_error(452,"write",err=errno);
            if (err == EAGAIN) errno = EIO;
            n = -1;
            break;
          }
          up += n;
        }
        if (n >= 0)
        { if (feof(sock)) n = 0; else
          { if (errno == EAGAIN)
            { up--;
              n = 1;
            } else n = -1;
          }
        }
      } else
      { n = read(fd_sock,buffer,buffersize);
        if (n < 0) send_error(452,"recv",err=errno); else
        if (n > 0)
        { ok = write(fd_file,buffer,n);
          if (ok < 0)
          { send_error(452,"write",err=errno);
            n = -1;
          } else if (ok < n)
          { send_error(452,"write",err=ENOSPC);
            n = -1;
          }
        }
      }
      if (n > 0)
      { up += n;
        /* enforce minimum and maximum speed, t2=time elapsed, t3=time limit */
        if (min_speed || max_speed)
        { t2 = time(NULL) - t1;
          if (min_speed && (t2 >= 30))
          { t3 = (up/1024) / min_speed;
            if (t2 > t3)
            { sprintf(buffer,"Minimum speed of %i Kb/s not reached, data connection closed",min_speed);
                send_reply(426,buffer);
                n = -1;
                break;
            }
          }
          if (max_speed)
          { t3 = (up/1024) / max_speed;
            if (t2 < t3)
            { t3 -= t2;
              sleep(t3 > 5 ? 5 : t3);
            }
          }
        }
      }
    }
  }
  fclose(file);
  fclose(sock);
  sock = NULL;
  time(&t2);
  t2 -= t1;
  t1 = t2;
  if (!t2) t2 = 1;
  t2 = last_speed = up / t2;
  if (!err)
  { /* upload succeeded */
  if ((ok = cmd_stor_check(buffer,dir,name)) == 2)
  { send_message(226,0,"Upload checker timeout. File not tested.",NULL);
    ok = 0;
  }
  } else
  { /* upload failed */
    ok = flag_get(flags,FLAG_NUKEBROKENUP);
  }
  if (!ok && limited)
  { bytes_uploaded += up - st->st_size;
    if (up < (unsigned int)st->st_size)
      printf("UUD -%u 0 %li 0\n",(unsigned int)st->st_size-up,t1);
      else printf("UUD %u 0 %li 0\n",up-(unsigned int)st->st_size,t1);
    scanf(SI64,&usr.credit);
    fgetc(stdin);
  }
  if (!n) st->st_size = -1; /* OK FLAG */
  maxfd = find_cdpath(dir) == NULL;
  *(name-1) = '/';
  add_dir(cwd,dirandname,buffer,buffersize);
  if (err) log_end(hidden,LG_PUT,"UPLOAD","%s FAILED: %s",buffer,strerror(err));
    else log_end(hidden,LG_PUT,"UPLOAD","%s (%lu bytes at %3.1f Kb/s)",
      buffer,up,(float)t2/1024.0);
  if (ok) return ok;
  if (maxfd)
  { printf("#1#FUP %s\n",buffer);
    fgets(buffer,buffersize,stdin);
  }
  return err ? 0 : send_message(226,1,upload_msg.msg,name);
}

int cmd_appe_stor(char** p, int n, int append)
{ char        buffer[4096];
  char*       dir;
  FILE*       file;
  FILE*       sock = NULL;
  struct stat st;
  char*       name;

  convert_filename(*p,fconv_mode & 0x0F);
  n = add_dir(cwd,*p,buffer,sizeof(buffer));
  if (n < 0) return send_error(550,"open",errno);
  if (append || (stat(buffer+1,&st) < 0)) st.st_size = 0;
  split_dir(*p,&dir,&name,0);
  if (strlen(name) >= MAXFILENAMELENGTH)
    return send_error(550,"open",ENAMETOOLONG);
  if ((permission_ok(dir,P_WRITE) == -1) || (!superuser() &&
    (!strcmp(name,readme_file))) || (!strcmp(name,permission_file)))
    return send_error(550,"write",EACCES);
  if (!append && find_dupe(name,buffer+2048))
    return send_message(550,1,"Dupe detected (%s)",buffer+2048);
  if (*pre_upload) switch (cmd_transferpermitted(pre_upload,buffer))
  { case -1 : return -1;
    case  1 : return 0;
  }
  switch (cmd_stor_dcheck(buffer,name))
  { case -1 : return -1;
    case  1 : return 0;
  }
  printf("LIM_GBW 1\n");
  if (!fgets(buffer,sizeof(buffer),stdin))
    return send_error(550,"pipe error",errno);
  if (*buffer == '+')
  { if (!sscanf(buffer+1,"%i",&n)) n = 0;
  } else n = 0;
  if (!n) return send_reply(550,"transfer not permitted (no bandwidth)");
  if (n < 0) n = 0;
  if (append)
    file = file_open(dir,name,type == T_ASCII ? "at" : "ab",P_WRITE); else
    file = file_open(dir,name,type == T_ASCII ? "wt" : "wb",P_WRITE);
  if (!file)
  { n = errno;
    printf("LIM_RBW\n");
    fgets(buffer,sizeof(buffer),stdin);
    return send_error(550,"write",n);
  }
  *(name-1) = '/';
  add_dir(cwd,*p,buffer,sizeof(buffer));
  if (!find_cdpath(buffer))
  { printf("#1#FCO "SI32" -2 %s\n",usr.uid,buffer);
    fgets(buffer,sizeof(buffer),stdin);
  }
  *(name-1) = '\0';
  n = cmd_store(buffer,sizeof(buffer),dir,name,*p,file,&sock,n,p,&st);
  if (st.st_size != -1)
  { fclose(file);
    if (flag_get(flags,FLAG_NUKEBROKENUP)) cmd_stor_rm(buffer,dir,name);
  }
  if (sock) fclose(sock);
  printf("LIM_RBW\n");
  fgets(buffer,sizeof(buffer),stdin);
  return n;
}

int cmd_stor(char** p, int n)
{ if (n != 1) return send_reply(501,"Syntax: STOR <filename>");
  return cmd_appe_stor(p,n,0);
}

int cmd_appe(char** p, int n)
{ if (n != 1) return send_reply(501,"Syntax: APPE <filename>");
  return cmd_appe_stor(p,n,1);
}

int cmd_port(char** p, int n, char* cmd, int cmdlen)
{ int  a1;
  int  a2;
  int  a3;
  int  a4;
  int  p1;
  int  p2;
  char str[32];

  if (n != 1) return send_reply(501,"Syntax: PORT <a1>,<a2>,<a3>,<a4>,<p1>,<p2>");
  rest_offset = 0;
  if (sscanf(*p,"%i,%i,%i,%i,%i,%i",&a1,&a2,&a3,&a4,&p1,&p2) != 6)
    return send_reply(501,"Syntax: PORT <a1>,<a2>,<a3>,<a4>,<p1>,<p2>");
  if (a1<0 || a1>255 || a2<0 || a2>255 || a3<0 || a3>255 || a4<0 || a4>255 ||
      p1<0 || p1>255 || p2<0 || p2>255)
    return send_reply(501,"Syntax: PORT <a1>,<a2>,<a3>,<a4>,<p1>,<p2>");
  sprintf(str,"%i.%i.%i.%i",a1,a2,a3,a4);
  memset(&addr,0,sizeof(addr));
  addr.sin_addr.s_addr = inet_addr(str);
  if (addr.sin_addr.s_addr == -1) return send_reply(501,"Invalid address.");
  p1 = (p1<<8)+p2;
  addr.sin_port = htons((u_short)p1);
  addr.sin_family = AF_INET;
  if (pasvsock != -1) close(pasvsock);
  pasvsock = -1;
  return send_reply(200,"PORT command successful.");
}

int cmd_pasv(char** p, int n, char* cmd, int cmdlen)
{ struct sockaddr_in addr;
  int                i;
  unsigned long      ipaddr;
  unsigned short     port;
  char               str[128];

  if (!flag_get(flags,FLAG_PASSIVEMODE))
    return send_reply(502,"Passive mode is disabled.");
  if (n) return send_reply(501,"Syntax: PASV");
  rest_offset = 0;
  if (pasvsock >= 0) close(pasvsock);
  pasvsock = -1;
  i = sizeof(addr);
  if (getsockname(sock,(struct sockaddr*)&addr,&i) < 0)
    return send_error(550,"getsockname",errno);
  pasvsock = socket(PF_INET,SOCK_STREAM,0);
  if (pasvsock < 0) return send_error(550,"socket",errno);
  addr.sin_family = AF_INET;
  addr.sin_port = 0;
  if (bind(pasvsock,(struct sockaddr*)&addr,sizeof(addr)) == -1)
  { i = errno;
    close(pasvsock);
    pasvsock = -1;
    return send_error(550,"bind",i);
  }
  if (listen(pasvsock,1) < 0)
  { i = errno;
    close(pasvsock);
    pasvsock = -1;
    return send_error(550,"listen",i);
  }
  i = sizeof(addr);
  if (getsockname(pasvsock,(struct sockaddr*)&addr,&i) == -1)
  { i = errno;
    close(pasvsock);
    pasvsock = -1;
    return send_error(550,"getsockname",i);
  }
  ipaddr = ntohl(addr.sin_addr.s_addr);
  port = ntohs(addr.sin_port);
  sprintf(str,"%lu,%lu,%lu,%lu,%hu,%hu",(ipaddr >> 24) & 0xFF,
   (ipaddr >> 16) & 0xFF,(ipaddr >> 8) & 0xFF,ipaddr & 0xFF,
   (port >> 8) & 0xFF,port & 0xFF);
  if (send_message(227,1,"Entering Passive Mode (%s).",str) < 0)
  { close(pasvsock);
    pasvsock = -1;
    return -1;
  }
  return 0;
}

/* EOF */
